"""
    Kucoin interface definition
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from blankly.exchanges.exchange import Exchange
from blankly.exchanges.auth.auth_constructor import AuthConstructor


class Kucoin(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "kucoin", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'kucoin', ['API_KEY', 'API_SECRET', 'API_PASS', 'sandbox'])

        sandbox = super().evaluate_sandbox(auth)

        try:
            from kucoin import client as KucoinAPI
        except ImportError:
            raise ImportError("Please \"pip install kucoin-python\" to use kucoin with blankly.")
        # Kucoin has a bunch of calls types, so we will index them in a single calls dictionary
        calls = {
            'market': KucoinAPI.Market(auth.keys['API_KEY'], auth.keys['API_SECRET'], auth.keys['API_PASS'], sandbox),
            'user': KucoinAPI.User(auth.keys['API_KEY'], auth.keys['API_SECRET'], auth.keys['API_PASS'], sandbox),
            'trade': KucoinAPI.Trade(auth.keys['API_KEY'], auth.keys['API_SECRET'], auth.keys['API_PASS'], sandbox)
        }

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    """
    Builds information about the asset on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self) -> dict:
        return self.calls




from blankly.exchanges.exchange import Exchange
from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.interfaces.okx.okx_api import Client as OkxAPI, MarketAPI, AccountAPI, TradeAPI, ConvertAPI, \
    FundingAPI, PublicAPI


class Okx(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "okx", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'okx', ['API_KEY', 'API_SECRET', 'API_PASS', 'sandbox'])
        keys = auth.keys

        sandbox = super().evaluate_sandbox(auth)

        # if self.preferences["settings"]["use_sandbox"]:
        calls = {
            'market': MarketAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'], passphrase=keys['API_PASS'],
                                sandbox=sandbox),
            'account': AccountAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'],
                                  passphrase=keys['API_PASS'],
                                  sandbox=sandbox),
            'trade': TradeAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'], passphrase=keys['API_PASS'],
                              sandbox=sandbox),
            'convert': ConvertAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'],
                                  passphrase=keys['API_PASS'],
                                  sandbox=sandbox),
            'funding': FundingAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'],
                                  passphrase=keys['API_PASS'],
                                  sandbox=sandbox),
            'public': PublicAPI(api_key=keys['API_KEY'], api_secret_key=keys['API_SECRET'], passphrase=keys['API_PASS'],
                                sandbox=sandbox)
        }

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    """
    Builds information about the asset on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self) -> dict:
        return self.calls



"""
    Coinbase Pro exchange definitions and setup
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_api import API as CoinbaseProAPI


class CoinbasePro(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "coinbase_pro", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'coinbase_pro', ['API_KEY', 'API_SECRET', 'API_PASS',
                                                                           'sandbox'])

        sandbox = super().evaluate_sandbox(auth)

        keys = auth.keys
        if sandbox:
            calls = CoinbaseProAPI(api_key=keys['API_KEY'],
                                   api_secret=keys['API_SECRET'],
                                   api_pass=keys['API_PASS'],
                                   api_url="https://api-public.sandbox.pro.coinbase.com/")
        else:
            # Create the authenticated object
            calls = CoinbaseProAPI(api_key=keys['API_KEY'],
                                   api_secret=keys['API_SECRET'],
                                   api_pass=keys['API_PASS'])

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    """
    Builds information about the asset on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self) -> CoinbaseProAPI:
        return self.calls

    """
    GUI Testing Functions | These only exist in the coinbase_pro class:
    """

    def get_indicators(self):
        return self.calls.get_products()

    def filter_relevant(self, dict):
        """
        Currently unused. Will be used for filtering for user-relevant currencies

        Args:
            dict: Dictionary to filter important currencies from

        Returns:
            list: filtered information about the account dictionary
        """
        self.get_state()
        self.__readable_state = {}
        unused = {}
        for i in range(len(self.__state)):
            value = float((self.__state[i]["balance"]))
            if value > 0:
                self.__readable_state[self.__state[i]["currency"]] = {
                    "Qty:": value,
                }
                currency = (self.__state[i]["currency"])
                # There can be no currency state until a model is added to the currency
                try:
                    self.__readable_state[self.__state[i]["currency"]] = {
                        **self.__readable_state[self.__state[i]["currency"]], **self.get_model_state(currency)}
                except KeyError as e:
                    pass
            else:
                unused[self.__state[i]["currency"]] = {
                    "Qty:": 0,
                }
        return self.__readable_state



"""
    FTX Futures impl of FuturesExchangeInterface.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import operator
from typing import Optional

from blankly.enums import MarginType, HedgeMode, Side, PositionMode, TimeInForce, ContractType, OrderStatus, OrderType
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
from blankly.exchanges.interfaces.ftx.ftx_interface import FTXInterface
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder
from blankly.utils import utils, time_builder
import datetime
import math


class FTXFuturesInterface(FuturesExchangeInterface):
    calls: FTXAPI

    @staticmethod
    def to_exchange_symbol(symbol: str):
        base_asset, quote_asset = symbol.split('-', 1)
        if quote_asset != 'USD':
            raise ValueError('invalid symbol')
        return base_asset + '-PERP'  # only perpetual contracts right now

    @staticmethod
    def to_blankly_symbol(symbol: str):
        base_asset, contract_type = symbol.split('-', 1)
        if contract_type != 'PERP':
            raise ValueError(
                'invalid symbol -- blankly only supports perpetual contracts')
        return base_asset + '-USD'

    @staticmethod
    def parse_timestamp(time: str) -> int:
        return int(datetime.datetime.fromisoformat(time).timestamp())

    @staticmethod
    def to_order_status(status: str, cancel: bool = False) -> OrderStatus:
        if status in ('new', 'open'):
            return OrderStatus.OPEN
        elif status == 'closed':
            if cancel:
                return OrderStatus.CANCELED
            else:
                return OrderStatus.FILLED
        raise ValueError(f'invalid order status: {status}')

    @staticmethod
    def to_order_type(type: str) -> OrderType:
        if type == 'market':
            return OrderType.MARKET
        elif type == 'limit':
            return OrderType.LIMIT
        elif type == 'takeProfit':
            return OrderType.TAKE_PROFIT
        elif type == 'stop':
            return OrderType.STOP
        raise ValueError(f'invalid order type: {type}')

    def parse_order_response(self,
                             response: dict,
                             cancel: bool = False) -> FuturesOrder:
        return FuturesOrder(
            symbol=response['future'],
            id=int(response['id']),
            status=self.to_order_status(response['status'], cancel),
            size=float(response['size']),
            type=self.to_order_type(response['type']),
            contract_type=ContractType.PERPETUAL,
            side=Side(response['side']),
            position=PositionMode.BOTH,
            price=float(response['avgFillPrice']),
            limit_price=float(response['price']),
            time_in_force=TimeInForce.IOC
            if response['ioc'] else TimeInForce.GTC,
            response=response,
            interface=self)

    def init_exchange(self):
        # will throw exception if our api key is stinky
        self.calls.get_account_info()

    def get_products(self, filter: str = None) -> dict:
        res = self.calls.list_futures()
        products = {}
        for prod in res:
            if '-PERP' not in prod['name']:
                continue  # only perpetual contracts for now
            symbol = prod['underlying'] + '-USD'
            products[symbol] = {
                'symbol': symbol,
                'base_asset': prod['underlying'],
                'quote_asset': 'USD',
                'contract_type': ContractType.PERPETUAL,
                'price_precision': utils.increment_to_precision(
                    prod['priceIncrement']),
                'size_precision': utils.increment_to_precision(
                    prod['sizeIncrement']),
                'exchange_specific': prod
            }
        if filter:
            return products[filter]
        return products

    def get_account(self, filter: str = None) -> dict:
        balances = self.calls.get_balances()
        coins = self.calls.get_coins()
        accounts = {
            coin['id']: {
                'available': 0.0,
                'hold': 0.0,  # TODO
                'exchange_specific': coin
            }
            for coin in coins
        }

        for bal in balances:
            coin = bal['coin']
            accounts[coin].available = float(bal['free'])
            # merge into exchange_specific
            accounts[coin].exchange_specific = {
                **accounts[coin].exchange_specific,
                **bal
            }

        if filter:
            return accounts[filter]
        return accounts

    def get_position(self, filter: str = None) -> Optional[dict]:
        leverage = self.get_leverage()

        res = self.calls.get_positions()
        positions = {}
        for position in res:
            symbol = self.to_exchange_symbol(position['future'])
            size = float(position['netSize'])
            if size == 0:
                continue
            positions[symbol] = {
                'size': size,
                'position': PositionMode.BOTH,
                'contract_type': ContractType.PERPETUAL,
                'exchange_specific': position
            }

        if filter:
            return positions.get(filter, None)
        return positions

    def market_order(self,
                     symbol: str,
                     side: Side,
                     size: float,
                     position: PositionMode = PositionMode.BOTH,
                     reduce_only: bool = False) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        if position != PositionMode.BOTH:
            raise ValueError(
                f'position mode {position} not supported on FTX Futures')
        res = self.calls.place_order(symbol, side, None, size, 'market',
                                     reduce_only)
        return self.parse_order_response(res)

    def limit_order(
            self,
            symbol: str,
            side: Side,
            price: float,
            size: float,
            position: PositionMode = PositionMode.BOTH,
            reduce_only: bool = False,
            time_in_force: TimeInForce = TimeInForce.GTC) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        if time_in_force == TimeInForce.GTC:
            ioc = False
        elif time_in_force == TimeInForce.IOC:
            ioc = True
        else:
            raise ValueError(
                f'time in force {time_in_force} not supported on FTX Futures')
        if position != PositionMode.BOTH:
            raise ValueError(
                f'position mode {position} not supported on FTX Futures')
        res = self.calls.place_order(symbol,
                                     side,
                                     price,
                                     size,
                                     'limit',
                                     reduce_only,
                                     ioc=ioc)
        return self.parse_order_response(res)

    def take_profit_order(
            self,
            symbol: str,
            side: Side,
            price: float,
            size: float,
            position: PositionMode = PositionMode.BOTH) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        if position != PositionMode.BOTH:
            raise ValueError(
                f'position mode {position} not supported on FTX Futures')
        res = self.calls.place_conditional_order(symbol,
                                                 side,
                                                 size,
                                                 'takeProfit',
                                                 trigger_price=price)
        return self.parse_order_response(res)

    def stop_loss_order(self,
                        symbol: str,
                        side: Side,
                        price: float,
                        size: float,
                        position: PositionMode = PositionMode.BOTH) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        if position != PositionMode.BOTH:
            raise ValueError(
                f'position mode {position} not supported on FTX Futures')
        res = self.calls.place_conditional_order(symbol,
                                                 side,
                                                 size,
                                                 'stop',
                                                 trigger_price=price)
        return self.parse_order_response(res)

    @utils.order_protection
    def set_hedge_mode(self, hedge_mode: HedgeMode):
        if hedge_mode == HedgeMode.HEDGE:
            raise Exception('HEDGE mode not supported on FTX Futures')
        pass  # FTX only has ONEWAY mode

    def get_hedge_mode(self):
        return HedgeMode.ONEWAY

    @utils.order_protection
    def set_leverage(self, leverage: float, symbol: str = None):
        if symbol:
            raise Exception(
                'FTX Futures does not allow setting leverage per symbol. Use interface.set_leverage(leverage) to set '
                'account-wide leverage instead. ')
        self.calls.change_account_leverage(leverage)

    def get_leverage(self, symbol: str = None) -> float:
        # doesn't matter if symbol passed or not, ftx leverage is same for all
        return self.calls.get_account_info()['leverage']

    @utils.order_protection
    def set_margin_type(self, symbol: str, type: MarginType):
        if type == MarginType.ISOLATED:
            raise Exception('isolated margin not supported on FTX Futures')
        pass

    def get_margin_type(self, symbol: str):
        return MarginType.CROSSED

    def cancel_order(self, symbol: str, order_id: int) -> FuturesOrder:
        res = self.get_order(symbol, order_id)
        res.status = OrderStatus.CANCELED
        self.calls.cancel_order(str(order_id))
        return res

    def get_open_orders(self, symbol: str = None) -> list:
        return [
            self.parse_order_response(o) for o in self.calls.get_open_orders()
        ]

    def get_order(self, symbol: str, order_id: int) -> FuturesOrder:
        response = self.calls.get_order_by_id(str(order_id))
        symbol = self.to_exchange_symbol(symbol)
        if response['symbol'] != symbol:
            raise Exception(
                'response symbol did not match parameter -- this should never happen'
            )
        return self.parse_order_response(response)

    def get_price(self, symbol: str) -> float:
        symbol = self.to_exchange_symbol(symbol)
        return float(self.calls.get_future(symbol)['mark'])

    def get_funding_rate_history(self, symbol: str, epoch_start: int,
                                 epoch_stop: int) -> list:
        symbol = self.to_exchange_symbol(symbol)
        # TODO dedup binance_futures_exchange maybe?
        history = []
        resolution = self.get_funding_rate_resolution()
        limit = 500
        window_start = epoch_start
        window_end = epoch_start + limit * resolution

        response = True
        while response:
            response = self.calls.get_funding_rates(window_start, window_end,
                                                    symbol)

            history.extend({
                               'rate': float(e['rate']),
                               'time': self.parse_timestamp(e['time'])
                           } for e in response)

            if response:
                window_start = window_end
                window_end = min(epoch_stop, window_start + limit * resolution)

        return sorted(history, key=operator.itemgetter('time'))

    def get_funding_rate_resolution(self) -> int:
        return time_builder.build_hour()

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        return FTXInterface.ftx_product_history(self.calls,self.to_exchange_symbol(symbol), epoch_start, epoch_stop, resolution)
    def get_maker_fee(self) -> float:
        raise NotImplementedError

    def get_taker_fee(self) -> float:
        raise NotImplementedError

    def get_funding_rate(self, symbol: str) -> float:
        raise NotImplementedError



from _pytest.python import Metafunc

from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface

from blankly.exchanges.futures.futures_exchange import FuturesExchange

from tests.new_interface_tests.test_utils import FUTURES_EXCHANGES, get_symbols, SPOT_EXCHANGES


def gen_id(arg):
    if isinstance(arg, ABCBaseExchangeInterface):
        return arg.get_exchange_type()
    return arg


def pytest_generate_tests(metafunc: Metafunc):
    interface_params = [p_name for p_name in metafunc.fixturenames if p_name.endswith('interface')]
    if len(interface_params) < 1:
        return  # no interface params is fine

    if len(interface_params) > 1:
        raise Exception('maximum one interface parameter per test function')

    interface_param = interface_params[0]

    if interface_param == 'futures_interface':
        interfaces = [ex.interface for ex in FUTURES_EXCHANGES]
    elif interface_param == 'spot_interface':
        interfaces = [ex.interface for ex in SPOT_EXCHANGES]
    elif interface_param == 'interface':
        interfaces = [ex.interface for ex in FUTURES_EXCHANGES + SPOT_EXCHANGES]
    else:
        raise Exception('invalid interface type: ' + interface_param)

    if 'symbol' in metafunc.fixturenames:
        vals = [(interface, symbol) for interface in interfaces
                for symbol in get_symbols(interface)]
        metafunc.parametrize(f'{interface_param},symbol', vals, ids=gen_id)
    else:
        metafunc.parametrize(interface_param, interfaces, ids=gen_id)



import blankly

if __name__ == "__main__":
    exchange = blankly.EXCHANGE_CLASS()
    strategy = blankly.Strategy(exchange)

    if blankly.is_deployed:
        strategy.start()
    else:
        strategy.backtest(to='1y', initial_values={'QUOTE_ASSET': 10000})



"""
    Base ExchangeInterface object.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
import time
import warnings
from collections import deque
from datetime import datetime as dt
from typing import Union

import numpy
import pandas
import pandas as pd

from blankly import utils
from blankly.utils import time_interval_to_seconds


# A lot of this class is just glue between ExchangeInterface and the new Futures classes.
# At some point it should probably all be refactored away but for now let's get futures working!
class ABCBaseExchangeInterface(abc.ABC):

    @abc.abstractmethod
    def get_exchange_type(self):
        pass

    @abc.abstractmethod
    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        pass

    def history(self,
                symbol: str,
                to: Union[str, int] = 200,
                resolution: Union[str, float] = '1d',
                start_date: Union[str, dt, float] = None,
                end_date: Union[str, dt, float] = None,
                return_as: str = 'df'):

        start, stop, res_seconds, to, present = self.calculate_epochs(start_date, end_date, resolution, to)

        response = self.overridden_history(symbol, start, stop, res_seconds, to=to,)

        # Add a check to make sure that coinbase pro has updated
        # I tried to delete this code but the entire function broke :(
        if present and self.get_exchange_type() == "coinbase_pro":
            data_append = None
            tries = 0
            while True:
                if data_append is None:
                    # We can continue if this is valid
                    if response['time'].iloc[-1] == stop:
                        break
                else:
                    if data_append[0]['time'] == stop:
                        break
                time.sleep(.5)
                tries += 1
                if tries > 5:
                    # Admit failure and return
                    warnings.warn("Exchange failed to provide updated data within the timeout.")
                    return self.cast_type(response, return_as)
                try:
                    data_append = [self.get_product_history(symbol,
                                                            stop - res_seconds,
                                                            stop,
                                                            res_seconds).iloc[-1].to_dict()]
                    data_append[0]['time'] = int(data_append[0]['time'])
                except IndexError:
                    # If there is no recent data on the exchange this will be an empty dataframe.
                    # This happens for low volume
                    utils.info_print("Most recent bar at this resolution does not yet exist - skipping.")
                    break

            response = response.append(data_append, ignore_index=True)

        # Determine the deque length - we really should use this generally
        if isinstance(to, int):
            point_count = to
        else:
            point_count = int((stop-start)/res_seconds + 1)
        # response.index = pd.to_datetime(response['time'], unit='s')
        return self.cast_type(response, return_as, point_count)

    def calculate_epochs(self, start_date, end_date, resolution, to):
        is_backtesting = self.backtesting_time()
        if is_backtesting is not None and end_date is None:
            # is_backtesting can only return a non None value if a function overrides the is_backtesting function
            #  and says its backtesting by returning a valid time
            end_date = is_backtesting
        if start_date is not None and end_date is not None:
            to = None
        to_present = False
        if end_date is None:
            to_present = True
        # convert resolution into epoch seconds
        resolution_seconds = int(time_interval_to_seconds(resolution))
        if end_date is None:
            # Figure out the next point and then subtract to the last stamp
            most_recent_valid_resolution = utils.ceil_date(dt.now(),
                                                           seconds=resolution_seconds).timestamp() - resolution_seconds
            # Binance is nice enough to update OHLCV data, so we have to exclude that by subtracting a resolution
            epoch_stop = most_recent_valid_resolution - resolution_seconds
            count_from = most_recent_valid_resolution
        else:
            if isinstance(end_date, str):
                from dateutil import parser
                parsed_date = parser.parse(end_date)
            elif isinstance(end_date, float) or isinstance(end_date, numpy.int64) or isinstance(end_date, int) or \
                    isinstance(end_date, numpy.int32):
                parsed_date = dt.fromtimestamp(end_date)
            else:
                parsed_date = end_date
            valid_time_in_past = utils.ceil_date(parsed_date,
                                                 seconds=resolution_seconds).timestamp() - resolution_seconds
            epoch_stop = valid_time_in_past - resolution_seconds
            count_from = valid_time_in_past
        if start_date is None and end_date is None:
            if isinstance(to, int):
                # use number of points to calculate the start epoch
                epoch_start = count_from - (to * resolution_seconds)
            else:
                epoch_start = count_from - time_interval_to_seconds(to)
        elif start_date is None and end_date is not None:
            if isinstance(to, int):
                epoch_start = count_from - (to * resolution_seconds)
            else:
                epoch_start = count_from - time_interval_to_seconds(to)
        else:
            epoch_start = utils.convert_input_to_epoch(start_date)
        return epoch_start, epoch_stop, resolution_seconds, to, to_present

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution, **kwargs) -> pd.DataFrame:
        return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def backtesting_time(self):
        return None

    @staticmethod
    def cast_type(response: pd.DataFrame, return_as: str, point_count=None):
        if return_as != 'df' and return_as != 'deque':
            return response.to_dict(return_as)
        elif return_as == 'deque':
            # Create a deque object that has the same length
            response = response.to_dict('list')
            for i in response.keys():
                response[i] = deque(response[i], point_count)
            return response
        elif return_as == 'df':
            return response
        else:
            utils.info_print(f"Return type {return_as} is not supported.")
        return response



"""
    Spot specific class for strategy
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import threading
import time
import traceback
import typing
import warnings

import blankly
from blankly.exchanges.abc_base_exchange import ABCBaseExchange
from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.exchanges.strategy_logger import StrategyLogger
from blankly.frameworks.model.model import Model
from blankly.frameworks.strategy.strategy_base import StrategyBase, EventType
from blankly.frameworks.strategy import StrategyState
from blankly.utils.utils import info_print


class StrategyStructure(Model):
    def __init__(self, exchange: Exchange):
        self.lock = threading.Lock()
        super().__init__(exchange)

        self.orderbook_websockets = None
        self.ticker_websockets = None
        self.ticker_manager = None
        self.orderbook_manager = None
        self.schedulers = None
        self.remote_backtesting = None

    def construct_strategy(self, schedulers, orderbook_websockets,
                           ticker_websockets, orderbook_manager, ticker_manager):
        self.schedulers = schedulers
        self.orderbook_websockets = orderbook_websockets
        self.ticker_websockets = ticker_websockets
        self.orderbook_manager = orderbook_manager
        self.ticker_manager = ticker_manager

    def rest_event(self, **event):
        callback = event['callback']  # type: callable
        symbol = event['symbol']  # type: str
        resolution = event['resolution']  # type: int
        variables = event['variables']  # type: dict
        type_ = event['type']  # type: EventType
        state = event['state']  # type: StrategyState

        state.variables = variables
        state.resolution = resolution

        if type_ == EventType.bar_event:
            if not self.is_backtesting:
                bar_time = event['bar_time']
                while True:
                    # Sometimes coinbase doesn't download recent data correctly
                    try:
                        # A few seconds should pass before querying alpaca because any solution is acceptable
                        if self.interface.get_exchange_type() == "alpaca":
                            time.sleep(2)
                            data = self.interface.history(symbol=symbol, to=1, resolution=resolution).iloc[-1].to_dict()
                            break
                        else:
                            data = self.interface.history(symbol=symbol, to=1, resolution=resolution).iloc[-1].to_dict()
                            if data['time'] + resolution == bar_time:
                                break
                    except IndexError:
                        pass
                    time.sleep(.5)
            else:
                # If we are backtesting always just grab the last point and hope for the best of course
                try:
                    data = self.interface.history(symbol=symbol, to=1, resolution=resolution).iloc[-1].to_dict()
                except IndexError:
                    warnings.warn("No bar found for this time range")
                    return

            args = [data, symbol, state]
        elif type_ == EventType.price_event:
            data = self.interface.get_price(symbol)
            args = [data, symbol, state]
        elif type_ == EventType.scheduled_event:
            args = [state]
        elif type_ == EventType.arbitrage_event:
            prices = {}
            # If we're backtesting loop through the symbol and just grab the price
            if self.is_backtesting:
                for sym in symbol:
                    prices[sym] = self.interface.get_price(sym)

            # We have to be a bit more strategy if we're live
            else:
                def grab_price(threaded_symbol):
                    prices[threaded_symbol] = self.interface.get_price(threaded_symbol)

                threadpool = []
                for sym in symbol:
                    threadpool.append(threading.Thread(target=grab_price, args=(sym,)))

                # Start the threads
                for thread in threadpool:
                    thread.start()

                # Join each of the threads because they've writen to the symbol
                for thread in threadpool:
                    thread.join()
            args = [prices, symbol, state]
        else:
            return

        try:
            callback(*args)
        except Exception:
            traceback.print_exc()

    def run_price_events(self, events: list):
        # run all events once at start
        for event in events:
            event['next_run'] = self.backtester.initial_time

        while self.has_data:
            events.sort(key=lambda d: d['next_run'])
            event = events[0]

            # Sleep the difference
            self.sleep(event['next_run'] - self.time)

            # Run the event
            next_run = self.rest_event(**event)
            if next_run:
                # if rest_event returns something, run this event again at that time
                # this implies the event did *not* run
                event['next_run'] = next_run
                event['was_delayed'] = True
            else:
                # otherwise, the event ran. we can revalue account and re-run normally @ `resolution` intervals
                self.backtester.value_account()
                event['next_run'] += event['resolution']

    def main(self, args):
        if self.is_backtesting:
            self.run_backtest()
        else:
            self.run_live()

    def run_backtest(self):
        # Write in the new interface, no matter which type it is
        for scheduler in self.schedulers:
            kwargs = scheduler.get_kwargs()
            # Overwrite the internal interface in the created strategy
            kwargs['state'].strategy.interface = self.interface
        self.__run_init()

        events = []
        for scheduler in self.schedulers:
            events.append(scheduler.get_kwargs())

        self.run_price_events(events)

    def __run_init(self):
        # Switch to live mode for the inits
        if self.is_backtesting:
            self.interface.backtesting = False
        for i in self.schedulers:
            kwargs = i.get_kwargs()
            if kwargs['init'] is not None:
                if kwargs['type'] != EventType.scheduled_event:
                    kwargs['init'](kwargs['symbol'], kwargs['state'])
                else:
                    kwargs['init'](kwargs['state'])

        # Switch back to the backtesting status
        self.interface.backtesting = self.is_backtesting

    def run_live(self):
        self.__run_init()

        for scheduler in self.schedulers:
            scheduler.start()

        for i in self.orderbook_websockets:
            # Index 2 contains the initialization function for the assigned websockets array
            if i[2] is not None:
                i[2](i[0], i[3])
            self.orderbook_manager.restart_ticker(i[0], i[1])

        for i in self.ticker_websockets:
            # The initialization function should have already been called for ticker websockets
            # Notice this is different from orderbook websockets because these are put into the scheduler
            self.ticker_manager.restart_ticker(i[0], i[1])

    def teardown(self):
        self.lock.acquire()
        for i in self.schedulers:
            i.stop_scheduler()
            kwargs = i.get_kwargs()
            teardown = kwargs['teardown']
            state_object = kwargs['state']
            symbol = kwargs['symbol']
            if callable(teardown):
                teardown(symbol, state_object)

        for i in self.orderbook_websockets:
            self.orderbook_manager.close_websocket(override_symbol=i[0], override_exchange=i[1])
            # Call the stored teardown
            teardown_func = i[4]
            if callable(teardown_func):
                teardown_func(i[3])

        for i in self.ticker_websockets:
            self.ticker_manager.close_websocket(override_symbol=i[0], override_exchange=i[1])
        self.lock.release()


class Strategy(StrategyBase):
    __exchange: Exchange
    interface: ABCExchangeInterface

    def __init__(self, exchange: ABCBaseExchange):
        self.model = StrategyStructure(exchange)
        super().__init__(exchange, StrategyLogger(exchange.get_interface(), strategy=self), model=self.model)
        self._paper_trade_exchange = blankly.PaperTrade(exchange)
        self.__prices_added = False

    def teardown(self):
        pass

    def backtest(self,
                 to: str = None,
                 initial_values: dict = None,
                 start_date: typing.Union[str, float, int] = None,
                 end_date: typing.Union[str, float, int] = None,
                 settings_path: str = None,
                 **kwargs
                 ) -> BacktestResult:
        """
        Turn this strategy into a backtest.

        Args:
            ** We expect either an initial_value (in USD) or a dictionary of initial values, we also expect
            either `to` to be set or `start_date` and `end_date` **

            to (str): Declare an amount of time before now to backtest from: ex: '5y' or '10h'
            initial_values (dict): Dictionary of initial value sizes (i.e { 'BTC': 3, 'USD': 5650}).
                Using this will override the values that are currently on your exchange.
            start_date (str): Override argument "to" by specifying a start date such as "03/06/2018". This can also
                be an epoch time as a float or int.
            end_date (str): End the backtest at a date such as "03/06/2018". This can also be an epoch type as a float
                or int
            settings_path (str): Path to the backtest.json file.

            Keyword Arguments:
                **Use these to override parameters in the backtest.json file**

                use_price: str = 'close',
                    Set which price column to use.

                smooth_prices: bool = False,
                    Create linear connections between downloaded prices

                GUI_output: bool = True,
                    Enable/disable GUI webpage display after backtest

                show_tickers_with_zero_delta: bool = False,
                    Exclude tickers that have no change to account value in the GUI

                save_initial_account_value: bool = True,
                    Put an extra frame which contains the initial account values before any trade

                show_progress_during_backtest: bool = True,
                    Show a progress bar as the backtest runs

                cache_location: str = './price_caches'
                    Set a location for the price cache csv's to be written to

                continuous_caching: bool
                    Utilize the advanced price caching system built into the backtest. Automatically aggregate and prune
                    downloaded data.

                resample_account_value_for_metrics: str or bool = '1d' or False
                    Because backtest data can be input at a variety of resolutions, account value often needs to be
                        recalculated at consistent intervals for use in metrics & indicators.
                        This setting allows the specification of that consistent interval.
                        The value can be set to `False` to skip any recalculation.

                quote_account_value_in: str = 'USD'
                    Manually set what valuation should be used when calculating account value.
                        Multiple types of quote currency (ex: USD and EUR) are not supported because
                        there is no datasource for quoting pairs such as EUR-USD until forex integration.

                ignore_user_exceptions: bool = True
                    Set this to True to handle user exceptions identically to how they're handled by strategy calls.
                        False means that the backtest will immediately stop & attempt to generate a report if something
                        in the user calls goes wrong. True will replicate strategy errors.

                risk_free_return_rate: float = 0.0
                    Set this to be the theoretical rate of return with no risk
        """
        self.setup_model()
        if len(self.orderbook_websockets) != 0 or len(self.ticker_websockets) != 0:
            info_print("Found websocket events added to this strategy. These are not yet backtestable without "
                       "event based data")

        self.__add_prices(to, start_date, end_date)
        res = self.model.backtest(args={}, initial_values=initial_values, settings_path=settings_path, kwargs=kwargs)
        self.model.teardown()
        return res

    def __add_prices(self, to, start_date, end_date):
        for scheduler in self.schedulers:
            event_element = scheduler.get_kwargs()

            # Skip any scheduled events
            if event_element['symbol'] is None:
                continue

            # Loop through the symbols if it is a list
            if isinstance(event_element['symbol'], list):
                for symbol in event_element['symbol']:
                    self.model.backtester.add_prices(to=to,
                                                     start_date=start_date,
                                                     stop_date=end_date,
                                                     symbol=symbol,
                                                     resolution=event_element['resolution'])
            else:
                self.model.backtester.add_prices(to=to,
                                                 start_date=start_date,
                                                 stop_date=end_date,
                                                 symbol=event_element['symbol'],
                                                 resolution=event_element['resolution'])

            self.__prices_added = True

        if not self.__prices_added:
            raise LookupError("No prices added. If using scheduled events, create an empty price event or add prices"
                              " manually using strategy.add_prices()")

    def add_prices(self, symbol: str, resolution: [str, int, float], to: str = None,
                   start_date: typing.Union[str, float, int] = None,
                   stop_date: typing.Union[str, float, int] = None):
        """
        Directly add prices to the strategy
        """
        self.model.backtester.add_prices(symbol, resolution, to, start_date, stop_date)
        self.__prices_added = True

    def setup_model(self):
        self.model.construct_strategy(self.schedulers, self.orderbook_websockets,
                                      self.ticker_websockets, self.orderbook_manager,
                                      self.ticker_manager)

    def start(self):
        """
        Run your model live!

        Simply call this function to take your strategy configuration live on your exchange
        """
        self.setup_model()
        if self.remote_backtesting:
            warnings.warn("Aborted attempt to start a live strategy a backtest configuration")
            return
        self.model.run()

    def time(self) -> float:
        return self.model.time



"""
    Websocket feeds need to be modular due to the subscription methods, this provides dynamic management for binance.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time

import blankly.utils.utils as utils


def switch_type(stream):
    if stream == "aggTrade":
        return trade, \
               trade_interface, \
               "event_time,system_time,event_type,symbol,trade_id,price,quantity," \
               "trade_time,buyer_is_maker\n"
    elif stream == "depth":
        return depth, depth_interface, ""
    else:
        return no_callback, no_callback, ""


def no_callback(message):
    return message


def depth(message):
    return message


def depth_interface(message):
    return message


def trade(message):
    return str(message["E"]) + "," + str(time.time()) + "," + message["e"] + "," + message["s"] + "," + \
           str(message["a"]) + "," + message["p"] + "," + message["q"] + "," + "," + "," + \
           str(message["T"]) + "," + str(message["m"]) + "\n"


def trade_interface(message):
    """
    Homogenizing with coinbase's return:
    {'type': 'ticker', 'sequence': 24587251151, 'symbol': 'BTC-USD', 'price': '56178.52', 'open_24h': '56881.78',
    'volume_24h': '17606.23228984', 'low_24h': '55288', 'high_24h': '58400', 'volume_30d': '506611.70878868',
    'best_bid': '56171.12', 'best_ask': '56178.53', 'side': 'sell', 'time': 1620331254.43236, 'trade_id': 165659167,
    'last_size': '0.04'}

    Response from trade streams
    {
      "e": "aggTrade",  // Event type
      "E": 123456789,   // Event time
      "s": "BNBBTC",    // Symbol
      "a": 12345,       // Aggregate trade ID
      "p": "0.001",     // Price
      "q": "100",       // Quantity
      "f": 100,         // First trade ID
      "l": 105,         // Last trade ID
      "T": 123456785,   // Trade time
      "m": true,        // Is the buyer the market maker?
      "M": true         // Ignore
    }

    Similar ticks with coinbase pro
    {
        'type': 'ticker',
        'symbol': 'BTC-USD',
        'price': '50141.55',
        'time': 1619286924.397969,
        'trade_id': 160775277,
        'size': .3432
    }
    """
    renames = [
        ["e", "type"],
        ["s", "symbol"],
        ["a", "trade_id"],
        ["p", "price"],
        ['q', "size"],
        ["T", "time"],
    ]
    message = utils.rename_to(renames, message)
    needed = [
        ["symbol", str],
        ["price", float],
        ["time", float],
        ["trade_id", int],
        ["size", float]
    ]
    message["time"] = message["time"] / 1000
    isolated = utils.isolate_specific(needed, message)
    isolated['symbol'] = utils.to_blankly_symbol(isolated['symbol'], 'binance')

    return isolated



"""
    Class to manage a limit order lifecycle - a way to determine individual information about each buy and sell
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from blankly.exchanges.orders.order import Order


class LimitOrder(Order):
    """
    (Buying or selling (string), amount in currency (BTC/XLM), ticker object (so we can get time and value),
    limit if there is one)
    """

    def __init__(self, order, response, interface):
        """
        Coinbase Response:
        {
            "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
            "price": "0.10000000",
            "size": "0.01000000",
            "product_id": "BTC-USD",
            "side": "buy",
            "stp": "dc",
            "type": "limit",
            "time_in_force": "GTC",
            "post_only": false,
            "created_at": "2016-12-08T20:02:28.53864Z",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "executed_value": "0.0000000000000000",
            "status": "pending",
            "settled": false
        }

        binance Response:
        {
            "symbol": "BTCUSDT",
            "orderId": 28,
            "clientOrderId": "6gCrw2kRUAF9CvJDGP16IP",
            "transactTime": 1507725176595,
            "price": "0.00000000",
            "origQty": "10.00000000",
            "executedQty": "10.00000000",
            "status": "FILLED",
            "timeInForce": "GTC",
            "type": "MARKET",
            "side": "SELL"
        }

        Guaranteed:
        needed = [
            ["symbol", str],
            ["id", str],
            ["created_at", float],
            ["price", float],
            ["size", float],
            ["status", str],
            ["time_in_force", str],
            ["type", str],
            ["side", str]
        ]
        """
        self.Interface = interface
        self.__exchange = self.Interface.get_exchange_type()
        self.__order = order
        self.__response = response
        self.__homogenized_result = None
        super().__init__(self.__response, self.__order, self.Interface)

    def get_price(self) -> float:
        """
        Get the price that the order was set at. For limits this will be user-specified, for markets this will
        be market price.
        """
        return self.__response["price"]

    def get_time_in_force(self) -> str:
        """
        Get the exchange's set time_in_force value.
        """
        return self.__response["time_in_force"]

    def __str__(self):
        return_string = super().__str__()

        return_string = self.add_new_line(return_string, "Limit Order Parameters: ")

        return_string = self.add_new_line(return_string, "Time In Force: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_time_in_force())

        return_string = self.add_new_line(return_string, "Price: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_price())

        return return_string

    """ This is something that should be implemented in the future """
    # """ Returns true if the buy order can be sold at a profit or the sell order didn't loose money """
    # def is_profitable(self, show=False):
    #     if self.__buyOrSell == "buy":
    #         if self.is_profitable_buy() < float(self.__ticker.get_most_recent_tick()["price"]):
    #             if show:
    #                 print(True)
    #             return True
    #         else:
    #             if show:
    #                 print(False)
    #             return False
    #     else:
    #         if self.is_profitable_buy() > float(self.__ticker.get_most_recent_tick()["price"]):
    #             if show:
    #                 print(True)
    #             return True
    #         else:
    #             if show:
    #                 print(False)
    #             return False
    #
    # def get_fee(self, show=False):
    #     fee = (self.__valueAtTime * self.__amountCurrency) * blankly.Constants.PRETEND_FEE_RATE
    #     if show:
    #         print(fee)
    #     return fee
    #
    # """
    # Returns the price that a BUY needs to reach to be profitable
    # Ths can be used to set limit orders
    # """
    # def is_profitable_buy(self, show=False):
    #     # price = self.__valueAtTime/(1 - Constants.PRETEND_FEE_RATE)
    #     # This one didn't work for the USD amount modified
    #     # price = self.__valueAtTime/(1 - (2 * Constants.PRETEND_FEE_RATE) + (Constants.PRETEND_FEE_RATE * Constants.PRETEND_FEE_RATE))
    #     maker_fee_rate = self.__exchange_properties['maker_fee_rate']
    #     price = ((maker_fee_rate + 1) * (self.__valueAtTime) / (1 - maker_fee_rate))
    #     if show:
    #         print(price)
    #     return price
    #
    # """
    # This allows this to be checked if this buy became profitable
    # """
    # def is_profitable_sell(self, show=False):
    #     self.__profitable = float((self.__ticker.get_most_recent_tick()["price"])) > self.is_profitable_buy()
    #     if show:
    #         print(self.__profitable)
    #     return self.__profitable



"""
    Logic to provide consistency across exchanges
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time

import pandas as pd

import blankly.utils.time_builder
import blankly.utils.utils as utils
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.exchanges.orders.stop_limit import StopLimit
from blankly.exchanges.orders.stop_loss import StopLossOrder
from blankly.exchanges.orders.take_profit import TakeProfitOrder
from blankly.utils.exceptions import APIException, InvalidOrder


class CoinbaseProInterface(ExchangeInterface):
    def __init__(self, exchange_name, authenticated_api):
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[60, 300, 900, 3600, 21600, 86400])

    def init_exchange(self):
        # This is purely an authentication check which can be disabled in settings
        # Coinbase is symbol agnostic so passing a blank here is okay
        fees = self.calls.get_fees()
        try:
            if fees['message'] == "Invalid API Key":
                raise LookupError("Invalid API Key - are you trying to use your normal exchange keys "
                                  "while in sandbox mode? \nTry toggling the \'sandbox\' setting "
                                  "in your keys.json or check if the keys were input correctly into your "
                                  "keys.json.")
        except KeyError:
            pass

    def get_products(self):
        needed = self.needed['get_products']
        """
        [
            {
                "id": "BTC-USD",
                "base_currency": "BTC",
                "quote_currency": "USD",
                "base_min_size": "0.0001",
                "base_max_size": "280",
                "quote_increment": "0.01",
                "base_increment": "0.00000001",
                "display_name": "BTC/USD",
                "min_market_funds": "5",
                "max_market_funds": "1000000",
                "margin_enabled": False,
                "post_only": False,
                "limit_only": False,
                "cancel_only": False,
                "trading_disabled": False,
                "status": "online",
                "status_message": "",
            },
        ]
        """
        products = self.calls.get_products()
        for i in range(len(products)):
            # Rename needed
            products[i]["symbol"] = products[i].pop("id")
            products[i]["base_asset"] = products[i].pop("base_currency")
            products[i]["quote_asset"] = products[i].pop("quote_currency")
            products[i] = utils.isolate_specific(needed, products[i])
        return products

    @utils.enforce_base_asset
    def get_account(self, symbol=None) -> utils.AttributeDict:
        """
        Get all currencies in an account, or sort by symbol/account_id
        Args:
            symbol (Optional): Filter by particular symbol

            These arguments are mutually exclusive
        Coinbase Pro: get_account
        """
        symbol = super().get_account(symbol=symbol)

        needed = self.needed['get_account']
        """
        [
            {
                "id": "71452118-efc7-4cc4-8780-a5e22d4baa53",
                "currency": "BTC",
                "balance": "0.0000000000000000",
                "available": "0.0000000000000000",
                "hold": "0.0000000000000000",
                "profile_id": "75da88c5-05bf-4f54-bc85-5c775bd68254"
            },
            {
                ...
            }
        ]
        """
        accounts = self.calls.get_accounts()

        parsed_dictionary = utils.AttributeDict({})

        # We have to sort through it if the accounts are none
        if symbol is not None:
            for i in accounts:
                if i["currency"] == symbol:
                    parsed_value = utils.isolate_specific(needed, i)
                    dictionary = utils.AttributeDict({
                        'available': parsed_value['available'],
                        'hold': parsed_value['hold']
                    })
                    return dictionary
            raise ValueError("Symbol not found")
        for i in range(len(accounts)):
            parsed_dictionary[accounts[i]['currency']] = utils.AttributeDict({
                'available': float(accounts[i]['available']),
                'hold': float(accounts[i]['hold'])
            })

        return parsed_dictionary

    @utils.order_protection
    def market_order(self, symbol, side, size) -> MarketOrder:
        """
        Used for buying or selling market orders
        Args:
            symbol: currency to buy
            side: buy/sell
            size: desired amount of base currency to buy or sell
        """
        needed = self.needed['market_order']
        """
        {
            'id': '8d3a6ea5-8d6d-4486-9c98-b266bae9a67c', 
            'product_id': 'BTC-USD', 
            'side': 'buy', 
            'stp': 'dc', 
            'funds': '39.92015968', 
            'specified_funds': '40', 
            'type': 'market', 
            'post_only': False, 
            'created_at': 1621015423.292914, 
            'fill_fees': '0', 
            'filled_size': '0', 
            'executed_value': '0', 
            'status': 'pending', 
            'settled': False
        }
        """
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = utils.build_order_info(0, side, size, symbol, 'market')
        response = self.calls.place_market_order(symbol, side, size=size)
        response = self._fix_response(needed, response)
        return MarketOrder(order, response, self)

    @utils.order_protection
    def limit_order(self, symbol, side, price, size) -> LimitOrder:
        """
        Used for buying or selling limit orders
        Args:
            symbol: currency to buy
            side: buy/sell
            price: price to set limit order
            size: amount of currency (like BTC) for the limit to be valued
        """
        needed = self.needed['limit_order']
        """
        {
            "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
            "price": "0.10000000",
            "size": "0.01000000",
            "product_id": "BTC-USD",
            "side": "buy",
            "stp": "dc",
            "type": "limit",
            "time_in_force": "GTC",
            "post_only": false,
            "created_at": "2016-12-08T20:02:28.53864Z",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "executed_value": "0.0000000000000000",
            "status": "pending",
            "settled": false
        }
        """
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = utils.build_order_info(price, side, size, symbol, 'limit')
        response = self.calls.place_limit_order(symbol, side, price, size=size)
        response = self._fix_response(needed, response)
        return LimitOrder(order, response, self)

    @staticmethod
    def _fix_response(needed, response):
        if "message" in response:
            raise InvalidOrder("Invalid Order: " + response["message"])
        response["created_at"] = utils.epoch_from_iso8601(response["created_at"])
        response["symbol"] = response.pop('product_id')
        response = utils.isolate_specific(needed, response)
        return response

    @utils.order_protection
    def take_profit_order(self, symbol, price, size) -> TakeProfitOrder:
        """
        Used for placing a take-profit orders
        Args:
            symbol: currency to sell
            price: price to sell at
            size: amount of currency (like BTC)
        """
        needed = self.needed['take_profit']
        """
        {
            "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
            "price": "0.10000000",
            "size": "0.01000000",
            "product_id": "BTC-USD",
            "side": "buy",
            "stp": "dc",
            "type": "limit",
            "time_in_force": "GTC",
            "post_only": false,
            "created_at": "2016-12-08T20:02:28.53864Z",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "executed_value": "0.0000000000000000",
            "status": "pending",
            "settled": false
        }
        """
        side = 'sell'
        order = utils.build_order_info(price, side, size, symbol, 'take_profit')
        response = self.calls.place_limit_order(symbol, side, price, size=size)
        response = self._fix_response(needed, response)
        return TakeProfitOrder(order, response, self)

    @utils.order_protection
    def stop_loss_order(self, symbol, price, size) -> StopLossOrder:
        """
        Used for placing a stop-loss order
        Args:
            symbol: currency to sell
            price: price to sell at
            size: amount of currency (like BTC)
        """
        needed = self.needed['stop_loss']
        """
        {
            "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
            "price": "0.10000000",
            "size": "0.01000000",
            "product_id": "BTC-USD",
            "side": "buy",
            "stp": "dc",
            "type": "limit",
            "time_in_force": "GTC",
            "post_only": false,
            "created_at": "2016-12-08T20:02:28.53864Z",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "executed_value": "0.0000000000000000",
            "status": "pending",
            "settled": false
        }
        """
        side = 'sell'
        order = utils.build_order_info(price, side, size, symbol, 'stop_loss')
        response = self.calls.place_stop_order(symbol, side, price=price, size=size)
        response = self._fix_response(needed, response)
        return StopLossOrder(order, response, self)

    """
    Stop limit isn't added to the abstract class because the binance version is barely supported.
    
    If you want to use this function you can, just do interface.stop_limit(args) if you're using a coinbase pro 
    interface. Keep in mind this only works live and is not integrated with paper trading or backtesting
    """

    def stop_limit(self, symbol, side, stop_price, limit_price, size, stop='loss') -> StopLimit:
        """
        Used for placing stop orders
        Args:
            symbol (str): currency to buy
            side (str): buy/sell
            stop_price (float): Price to trigger the stop order at
            limit_price (float): Price to set the stop limit
            size (float): Amount of base currency to buy: (1.3 BTC)
            stop (str) (optional): Stop type of "loss" or "entry"

        {
            'id': '3a98a5c6-05a0-4e46-b8e4-3f27358fe27d',
            'price': '29500',
            'size': '0.01',
            'product_id':
            'BTC-USD',
            'side': 'buy',
            'stp': 'dc',
            'type': 'limit',
            'time_in_force':
            'GTC',
            'post_only': False,
            'created_at': '2021-05-28T19:24:52.010449Z',
            'stop': 'loss',
            'stop_price': '30000',
            'fill_fees': '0',
            'filled_size': '0',
            'executed_value': '0',
            'status': 'pending',
            'settled': False
        }
        """
        order = {
            'symbol': symbol,
            'side': side,
            'type': 'stop_loss',
            'stop': stop,
            'stop_price': stop_price,
            'size': size,
            'price': limit_price
        }
        response = self.calls.place_order(product_id=symbol,
                                          order_type='limit',
                                          side=side,
                                          stop=stop,  # loss
                                          stop_price=stop_price,
                                          size=size,
                                          price=limit_price
                                          )
        response['limit_price'] = response.pop('price')
        response["symbol"] = response.pop('product_id')
        return StopLimit(order, response, self)

    def cancel_order(self, symbol, order_id) -> dict:
        """
        Returns:
            dict: Containing the order_id of cancelled order. Example::
            { "order_id": "c5ab5eae-76be-480e-8961-00792dc7e138" }
        """
        return {"order_id": self.calls.cancel_order(order_id)}

    def get_open_orders(self, symbol=None):
        """
        List open orders.
        """
        """
        [
            {
                "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
                "price": "0.10000000",
                "size": "0.01000000",
                "product_id": "BTC-USD",
                "side": "buy",
                "stp": "dc",
                "type": "limit",
                "time_in_force": "GTC",
                "post_only": false,
                "created_at": "2016-12-08T20:02:28.53864Z",
                "fill_fees": "0.0000000000000000",
                "filled_size": "0.00000000",
                "executed_value": "0.0000000000000000",
                "status": "open",
                "settled": false
            },
            {
                ...
            }
        ]
        """
        if symbol is None:
            orders = list(self.calls.get_orders())
        else:
            orders = list(self.calls.get_orders(product_id=symbol))

        if len(orders) == 0:
            return []
        if orders[0] == 'message':
            raise InvalidOrder("Invalid Order: " + str(orders))

        for i in range(len(orders)):
            orders[i]["created_at"] = utils.epoch_from_iso8601(orders[i]["created_at"])
            needed = self.choose_order_specificity(orders[i]['type'])
            orders[i]["symbol"] = orders[i].pop('product_id')
            orders[i] = utils.isolate_specific(needed, orders[i])

        return orders

    def get_order(self, symbol, order_id) -> dict:
        """
        {
            "created_at": "2017-06-18T00:27:42.920136Z",
            "executed_value": "0.0000000000000000",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "id": "9456f388-67a9-4316-bad1-330c5353804f",
            "post_only": true,
            "price": "1.00000000",
            "product_id": "BTC-USD",
            "settled": false,
            "side": "buy",
            "size": "1.00000000",
            "status": "pending",
            "stp": "dc",
            "time_in_force": "GTC",
            "type": "limit"
        }
        This is the response for a market order
        {
            'id': '8d3a6ea5-8d6d-4486-9c98-b266bae9a67c', 
            'side': 'buy', 
            'type': 'market', 
            'status': 'done',
            'product_id': 'BTC-USD',
            'profile_id': '4107a8cc-9fee-4ed4-b609-d87011675ed5',
            'funds': '39.9201596800000000', 
            'specified_funds': '40.0000000000000000',
            'post_only': False, 
            'created_at': '2021-05-14T18:03:43.292914Z',
            'done_at': '2021-05-14T18:03:43.297Z', 
            'done_reason': 'filled',
            'fill_fees': '0.0798400768200000', 
            'filled_size': '0.00078825',
            'executed_value': '39.9200384100000000', 
            'settled': True
        }
        """
        response = self.calls.get_order(order_id)
        if 'message' in response:
            # This part will run through all orders if the user enables the setting
            # Leaving this commented because its useful in getting all orders if we add that
            # if self.user_preferences['settings']['coinbase_pro']['search_all_fills']:
            #     all_orders = list(self.calls.get_orders(product_id=currency_id, status='all'))
            #     found = False
            #     for i in all_orders:
            #         if i['id'] == order_id:
            #             found = True
            #             response = i
            #
            #     if not found:
            #         raise APIException("Invalid: " + str(response['message']) + ", was the order canceled?")
            # else:
            raise APIException("Invalid: " + str(response['message']) + ", was the order canceled?")
        response["created_at"] = utils.epoch_from_iso8601(response["created_at"])

        if response['type'] == 'market':
            needed = self.needed['market_order']
        elif response['type'] == 'limit':
            needed = self.needed['limit_order']
        else:
            needed = self.needed['market_order']

        response["symbol"] = response.pop('product_id')
        return utils.isolate_specific(needed, response)

    """
    Coinbase Pro: get_fees
    binance: get_trade_fee
    """

    def get_fees(self, symbol) -> dict:
        needed = self.needed['get_fees']
        """
        {
            'maker_fee_rate': '0.0050',
            'taker_fee_rate': '0.0050',
            'usd_volume': '37.69'
        }
        """
        fees = self.calls.get_fees()
        return utils.isolate_specific(needed, fees)

    """
    Coinbase Pro: get_account_history
    binance: 
        get_deposit_history
        get_withdraw_history

    """

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        """
        Returns the product history from an exchange
        Args:
            symbol: blankly product ID format (BTC-USD)
            epoch_start: Time to begin download
            epoch_stop: Time to stop download
            resolution: Resolution in seconds between tick (ex: 60 = 1 per minute)
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
        """

        resolution = blankly.time_builder.time_interval_to_seconds(resolution)

        # epoch_start, epoch_stop = super().get_product_history(symbol, epoch_start, epoch_stop, resolution)
        epoch_start = utils.convert_epochs(epoch_start)
        epoch_stop = utils.convert_epochs(epoch_stop)

        accepted_grans = [60, 300, 900, 3600, 21600, 86400]
        if resolution not in accepted_grans:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = accepted_grans[min(range(len(accepted_grans)),
                                            key=lambda i: abs(accepted_grans[i] - resolution))]

        resolution = int(resolution)

        # Figure out how many points are needed
        need = int((epoch_stop - epoch_start) / resolution)
        initial_need = need
        window_open = epoch_start
        history = []
        # Iterate while its more than max
        while need > 300:
            # Close is always 300 points ahead
            window_close = window_open + 300 * resolution
            open_iso = utils.iso8601_from_epoch(window_open)
            close_iso = utils.iso8601_from_epoch(window_close)
            response = self.calls.get_product_historic_rates(symbol, open_iso, close_iso, resolution)
            if isinstance(response, dict):
                raise APIException(response['message'])
            history = history + response

            window_open = window_close
            need -= 300
            time.sleep(.2)
            utils.update_progress((initial_need - need) / initial_need)

        # Fill the remainder
        open_iso = utils.iso8601_from_epoch(window_open)
        close_iso = utils.iso8601_from_epoch(epoch_stop)
        response = self.calls.get_product_historic_rates(symbol, open_iso, close_iso, resolution)
        if isinstance(response, dict):
            raise APIException(response['message'])
        history_block = history + response
        history_block.sort(key=lambda x: x[0])

        df = pd.DataFrame(history_block, columns=['time', 'low', 'high', 'open', 'close', 'volume'])
        # df[['time']] = df[['time']].astype(int)
        # Have to cast this for some reason
        df[['time']] = df[['time']].astype(int)
        df[['low', 'high', 'open', 'close', 'volume']] = df[['low', 'high', 'open', 'close', 'volume']].astype(float)

        return df

    """
    Coinbase Pro: Get Currencies
    binance: get_products
    """

    def get_order_filter(self, symbol: str):
        """
        Returns:
        list: List of available currency pairs. Example::
            [
                {
                    "id": "BTC-USD",  <-- Needed
                    "display_name": "BTC/USD",
                    "base_currency": "BTC", <-- Needed
                    "quote_currency": "USD",  <-- Needed
                    "base_increment": "0.00000001",  <-- Needed
                    "quote_increment": "0.01000000",  <-- Needed
                    "base_min_size": "0.00100000",  <-- Needed
                    "base_max_size": "280.00000000",  <-- Needed
                    "min_market_funds": "5",
                    "max_market_funds": "1000000",
                    "status": "online",
                    "status_message": "",
                    "cancel_only": false,
                    "limit_only": false,
                    "post_only": false,
                    "trading_disabled": false
                },
                ...
            ]
            """
        response = self.calls.get_products()
        products = None
        for i in response:
            if i["id"] == symbol:
                products = i
                break

        if products is None:
            raise LookupError("Specified market not found")

        base_min_size = float(products.pop('base_increment'))
        # Coinbase removed base_max_size
        base_max_size = float(100000000000000000)
        base_increment = float(base_min_size)

        min_market_funds = products['min_market_funds']
        max_market_funds = 1000000000 # big number approach

        return {
            "symbol": products.pop('id'),
            "base_asset": products.pop('base_currency'),
            "quote_asset": products.pop('quote_currency'),
            "max_orders": 1000000000000,
            "limit_order": {
                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment
                # for the base_asset.
                "price_increment": float(products['quote_increment']),

                "min_price": float(products['quote_increment']),
                "max_price": 9999999999,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment

                "quote_increment": float(products.pop('quote_increment')),  # Specifies the min order price as well
                # as the price increment.
                "buy": {
                    "min_funds": float(min_market_funds),
                    "max_funds": float(max_market_funds),
                },
                "sell": {
                    "min_funds": float(min_market_funds),
                    "max_funds": float(max_market_funds),
                },
            },
            "exchange_specific": {**products}
        }

    def get_price(self, symbol) -> float:
        """
        Returns just the price of a currency pair.
        """
        """
        {
            'trade_id': 28976297, 
            'price': '57579.72', 
            'size': '0.00017367', 
            'time': '2021-05-01T17:01:32.717482Z', 
            'bid': '57579.72', 
            'ask': '57579.74', 
            'volume': '31137.51184419'
        }
        """
        response = self.calls.get_product_ticker(symbol)
        if 'message' in response:
            raise APIException("Error: " + response['message'])
        return float(response['price'])



"""
    Backtesting controller for paper trading interface
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import os
import time
import traceback
import typing
from datetime import datetime as dt
import copy
import enum
import blankly

import numpy as np
import pandas as pd
import requests
from bokeh.layouts import column as bokeh_columns
from bokeh.models import HoverTool
from bokeh.palettes import Category10_10
from bokeh.plotting import ColumnDataSource, figure, show

import blankly.exchanges.interfaces.paper_trade.metrics as metrics
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.exchanges.interfaces.paper_trade.futures.futures_paper_trade_interface import FuturesPaperTradeInterface
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import load_backtest_preferences, write_backtest_preferences, info_print, update_progress, \
    get_base_asset, get_quote_asset, aggregate_prices_by_resolution
from blankly.exchanges.interfaces.paper_trade.backtest.format_platform_result import \
    format_platform_result

from blankly.exchanges.interfaces.paper_trade.abc_backtest_controller import ABCBacktestController
from blankly.exchanges.exchange import ABCExchange
from blankly.data.data_reader import PriceReader, TickReader, DataReader, FundingRateEventReader


def to_string_key(separated_list):
    output = ""
    for i in range(len(separated_list) - 1):
        output += str(separated_list[i])
        output += ","
    output += str(separated_list[-1:][0])
    return output


def split(base_range, local_segments) -> typing.Tuple[list, list]:
    """
    Find the negative given from a range and a set of other ranges

    Args:
        base_range: Backing array containing a range such as [1, 10]
        local_segments: An array of segments that we have such as [-5, 5], [6,7]
    Returns:
        The output of the example inputs above would be [[8, 10]]
    """

    # If we don't have any local segments there is no need for any of this, just download the whole set
    used_ranges = []
    positive_ranges = []  # These are the ranges that we have downloaded
    negative_ranges = []  # These are the ranges that we need

    def intersection(as_, ae, bs, be):
        """
        Find the intersection of two ranges
        Implementation from https://scicomp.stackexchange.com/a/26260

        Args:
            as_: Low of first range
            ae: High of first range
            bs: Low of second range
            be: High of second range
        """
        if bs > ae or as_ > be:
            return None
        else:
            os_ = max(as_, bs)
            oe = min(ae, be)
            return [os_, oe]

    for i in local_segments:
        intersection_range = intersection(base_range[0], base_range[1], i[0], i[1])
        if intersection_range is None:
            continue
        # We now know every single downloaded range that we need to keep
        used_ranges.append(i)

        # Now these are all intersections that we will filter to find the negative ranges to download
        positive_ranges.append(intersection_range)

    # Create a list of indexes that need to be pushed into a single list, arrays like  [1, 5] and [3, 7] would need to
    #  become [1, 7]

    # First we have to sort the lists to have the same starting value
    positive_ranges = sorted(positive_ranges, key=lambda x: x[0])

    aggregate_indexes = []
    for i in range(len(positive_ranges) - 1):
        intersection_range = intersection(positive_ranges[i][0], positive_ranges[i][1],
                                          positive_ranges[i + 1][0], positive_ranges[i + 1][1])
        if intersection_range is not None:
            aggregate_indexes.append([i, i + 1])

    # Now aggregate from the last result
    for i in aggregate_indexes:
        # Pull the lower bound right out of the lower index from the aggregate_indexes approach
        lower_bound = positive_ranges[i[0]][0]

        positive_ranges[i[0]] = None

        # Push that same lower bound into the array that was above it
        positive_ranges[i[1]][0] = lower_bound

    # Filter out the None's that were added
    positive_ranges = [x for x in positive_ranges if not (x is None)]

    # Now just try to find the gaps in the positive ranges and those are our negative ranges
    for i in range(len(positive_ranges) - 1):
        negative_ranges.append([positive_ranges[i][1], positive_ranges[i + 1][0]])

    try:
        # Now we just have to make sure to check the bounds
        # If the very first value we have is greater than what we requested
        if positive_ranges[0][0] > base_range[0]:
            # We then need the base range up to our first positive value
            negative_ranges.append([base_range[0], positive_ranges[0][0]])

        # If the last positive range is less than our requested final
        if positive_ranges[-1][1] < base_range[1]:
            # The actual value is from the last positive range value to our requested final value
            negative_ranges.append([positive_ranges[-1][1], base_range[1]])
    except IndexError:
        return [], [base_range]

    return used_ranges, negative_ranges


class BackTestController(ABCBacktestController):  # circular import to type model
    def __init__(self, model):
        self.backtesting = False
        self.preferences = None
        self.backtest_settings_path = None

        self.interface = None

        self.initial_time = None
        self.model = model

        self.traded_account_values = []
        self.no_trade_account_values = []

        # Prices sorted by symbol and then records of prices
        self.prices = {}
        # A list of events sorted by time. All events are put into this single list
        self.events = []

        # User added times
        self.__user_added_times = []

        # Open, high, low, close or volume
        self.use_price = None

        # Should we write to the preferences?
        self.queue_backtest_write = False

        # The quote currency to use
        self.quote_currency = None

        # Create a global generator because a second yield function gets really nasty
        # This is used for the colors of the graphs
        self.__color_generator = Category10_10.__iter__()

        # Some initial account value to store globally
        self.initial_account = None

        # Create our own traded assets' dictionary because we customize it a bit
        self.__traded_assets = []

        # Export a time for use in other classes
        self.time = None

        # The smallest amount of time to save
        self.min_resolution = None

        # Set these when prices are added to find the first price and the last price
        self.user_start = None
        self.user_stop = None

        self.show_progress = False
        self.sleep_count = 0

        # Use this global to retain where we are in the prices dictionary by index
        self.price_indexes = {}
        # Use this to keep trace globally of the event index we're using
        self.event_index = 0

        # Custom injected price readers and events readers
        self.__price_readers = []
        self.__event_readers = []
        self.__tick_readers = []

    class PriceIdentifiers(enum.Enum):
        exchange: str = 0
        sandbox: bool = 1
        symbol: str = 2
        epoch_start: int = 3
        epoch_stop: int = 4
        resolution: int = 5

    def parse_events(self):
        """
        Transform to:
        [
            {
                "type": "news event",
                "data": "gotem",
                "time": 2
            }
        ]
        """
        # TODO some code duplication here for the different events
        for reader in self.__event_readers:
            # Get the data as dict of dataframes
            data = reader.data
            for event_type in data:
                self.__check_user_time_bounds(reader.data[event_type]['time'].iloc[0],
                                              reader.data[event_type]['time'].iloc[-1],
                                              60)
                # Get the dataframe in each one
                event_df: pd.DataFrame = data[event_type]
                # Now turn it into a set of records
                records = event_df.to_dict(orient='records')
                # And make sure to add on the event type to each one
                for record_index in range(len(records)):
                    records[record_index]['type'] = event_type

                self.events += records

        for tick_reader in self.__tick_readers:
            data = tick_reader.data
            for symbol in data:
                self.__check_user_time_bounds(tick_reader.data[symbol]['time'].iloc[0],
                                              tick_reader.data[symbol]['time'].iloc[-1],
                                              60)
                symbol_df: pd.DataFrame = data[symbol]

                records = symbol_df.to_dict(orient='records')
                data_formatted_records = []
                for record_index in range(len(records)):
                    data_formatted_records.append({
                        'type': '__blankly__tick',
                        'data': records[record_index],
                        'time': records[record_index]['time']
                    })

                self.events += data_formatted_records

        # Now we just need to sort by time
        self.events = sorted(self.events, key=lambda d: d['time'])

    def sync_prices(self) -> dict:
        """
        Parse the local file cache for the requested data, if it doesn't exist, request it from the exchange

        args:
            items: list of lists organized as ['symbol', 'start_time', 'end_time', 'resolution']

        returns:
            dictionary with keys for each 'symbol'
        """

        # Make sure the cache folder exists and read files
        cache_folder = self.preferences['settings']["cache_location"]

        def sort_prices_by_resolution(price_dict):
            for symbol_ in price_dict:
                for resolution_ in price_dict[symbol_]:
                    price_dict[symbol_][resolution_] = price_dict[symbol_][resolution_].sort_values(by=['time'],
                                                                                                    ignore_index=True)

            return price_dict

        def parse_identifiers() -> list:
            try:
                files = os.listdir(cache_folder)
            except FileNotFoundError:
                files = []
                os.mkdir(cache_folder)

            identifiers_ = []
            for file in range(len(files)):
                # example file name: 'coinbase_pro.sandbox.BTC-USD.1622400000.1622510793.60.csv'
                # Remove the .csv from each of the files: BTC-USD.1622400000.1622510793.60
                identifier = files[file][:-4].split(",")
                # Cast to float first before
                try:
                    identifiers_.append({
                        self.PriceIdentifiers.exchange: identifier[0],
                        self.PriceIdentifiers.sandbox: identifier[1] == 'True',
                        self.PriceIdentifiers.symbol: identifier[2],
                        self.PriceIdentifiers.epoch_start: int(float(identifier[3])),
                        self.PriceIdentifiers.epoch_stop: int(float(identifier[4])),
                        self.PriceIdentifiers.resolution: int(float(identifier[5]))
                    })
                except IndexError:
                    # Remove each of the failed cache objects
                    os.remove(os.path.join(cache_folder, files[file]))

            return identifiers_

        def sort_identifiers(identifiers_: list) -> dict:
            """
            Create the hierarchy based on the identifiers
            {
                'coinbase_pro': {
                    True: {
                        'BTC-USD': [60, 3600, 86400]
                    }
                    False: {

                    }
                }
            }
            """
            local_history_blocks_ = {}

            def sort_identifier(identifier_: dict):
                exchange_ = identifier_[self.PriceIdentifiers.exchange]
                sandbox_ = identifier_[self.PriceIdentifiers.sandbox]
                symbol_ = identifier_[self.PriceIdentifiers.symbol]
                resolution_ = identifier_[self.PriceIdentifiers.resolution]

                if exchange_ not in local_history_blocks_:
                    local_history_blocks_[exchange_] = {}

                if sandbox_ not in local_history_blocks_[exchange_]:
                    local_history_blocks_[exchange_][sandbox_] = {}

                if symbol_ not in local_history_blocks_[exchange_][sandbox_]:
                    local_history_blocks_[exchange_][sandbox_][symbol_] = {}

                # Now we're checking the resolution array
                if resolution_ not in local_history_blocks_[exchange_][sandbox_][symbol_]:
                    local_history_blocks_[exchange_][sandbox_][symbol_][resolution_] = []

                # Add these in as a dictionary
                local_history_blocks_[exchange_][sandbox_][symbol_][resolution_].append({
                    self.PriceIdentifiers.epoch_start: identifier_[self.PriceIdentifiers.epoch_start],
                    self.PriceIdentifiers.epoch_stop: identifier_[self.PriceIdentifiers.epoch_stop]
                })

            # This is only the downloaded data
            for identifier in identifiers_:
                sort_identifier(identifier)

            return local_history_blocks_

        identifiers = parse_identifiers()

        local_history_blocks = sort_identifiers(identifiers)

        final_prices: dict = {}
        prices_by_resolution: dict = {}
        for i in range(len(self.__user_added_times)):
            if self.__user_added_times[i] is None:
                continue
            symbol = self.__user_added_times[i][self.PriceIdentifiers.symbol]
            resolution = self.__user_added_times[i][self.PriceIdentifiers.resolution]
            start_time = self.__user_added_times[i][self.PriceIdentifiers.epoch_start]
            end_time = self.__user_added_times[i][self.PriceIdentifiers.epoch_stop] - resolution
            exchange = self.interface.get_exchange_type()
            sandbox = True

            if end_time < start_time:
                raise RuntimeError("Must specify a longer timeframe to run the backtest.")

            downloaded_ranges = []

            # Attempt to find the same symbol/asset possibilities in the backtest blocks
            try:
                # Make sure to copy it because if you don't you delete this for any similar resolution
                # If you don't copy it fails if you use two price events at the same resolution
                available_prices = copy.deepcopy(local_history_blocks[exchange][sandbox][symbol][resolution])
                # Pull the epoch start and epoch stop for this particular symbol / resolution
                for price_set in available_prices:
                    downloaded_ranges.append([
                        price_set[self.PriceIdentifiers.epoch_start],
                        price_set[self.PriceIdentifiers.epoch_stop]
                    ])
            except KeyError:
                pass

            used_ranges, negative_ranges = split([start_time, end_time], downloaded_ranges)

            relevant_data = []
            for j in used_ranges:
                relevant_data.append(pd.read_csv(os.path.join(cache_folder, to_string_key([exchange,
                                                                                           True,
                                                                                           symbol,
                                                                                           j[0],
                                                                                           j[1],
                                                                                           resolution]) + ".csv")))

            if len(relevant_data) > 0:
                final_prices[symbol] = pd.concat(relevant_data)
                for dataset in relevant_data:
                    prices_by_resolution = aggregate_prices_by_resolution(prices_by_resolution, symbol, resolution,
                                                                          dataset)

            # If there is any data left to download do it here
            for j in negative_ranges:
                print("No cached data found for " + symbol + " from: " + str(j[0]) + " to " +
                      str(j[1]) + " at a resolution of " + str(resolution) + " seconds.")
                download = self.interface.get_product_history(symbol,
                                                              j[0],
                                                              j[1],
                                                              resolution)

                # Write the file but this time include very accurately the start and end times
                if self.preferences['settings']['continuous_caching']:
                    if not download.empty:
                        download.to_csv(os.path.join(cache_folder, f'{exchange},'
                                                                   f'{True},'
                                                                   f'{symbol},'
                                                                   f'{int(j[0])},'
                                                                   f'{int(j[1]) + resolution},'  # This adds resolution 
                                                                                                 # back to the exported
                                                                                                 # time series
                                                                   f'{resolution}.csv'),
                                        index=False)

                prices_by_resolution = aggregate_prices_by_resolution(prices_by_resolution, symbol, resolution,
                                                                      download)
                # Write these into the data array
                if symbol not in final_prices:
                    final_prices[symbol] = download
                else:
                    final_prices[symbol] = pd.concat([final_prices[symbol], download])

            # After all the negative ranges are appended, we need to sort & trim
            final_prices[symbol] = final_prices[symbol].sort_values(by=['time'], ignore_index=True)

            # Now make sure to just trim our times to hit the start and end times
            final_prices[symbol] = final_prices[symbol][final_prices[symbol]['time'] >= start_time]
            final_prices[symbol] = final_prices[symbol][
                final_prices[symbol]['time'] <= end_time + resolution]  # Add back

        # Now add any custom prices
        for price_reader in self.__price_readers:
            data = price_reader.data
            for symbol in data:
                symbol_info = price_reader.prices_info[symbol]
                resolution = symbol_info['resolution']
                start_time = symbol_info['start_time']
                stop_time = symbol_info['stop_time']

                # Add each symbol to the final prices without doing any processing
                if symbol in final_prices:
                    final_prices[symbol] = pd.concat([final_prices[symbol], data[symbol]])
                else:
                    final_prices[symbol] = data[symbol]

                prices_by_resolution = aggregate_prices_by_resolution(prices_by_resolution, symbol, resolution,
                                                                      data[symbol])

                self.__check_user_time_bounds(start_time,
                                              stop_time,
                                              resolution)

        # Send the prices by resolution to the interface
        self.interface.receive_price_cache(sort_prices_by_resolution(prices_by_resolution))

        # Finally, convert back into records
        for symbol in final_prices:
            final_prices[symbol] = final_prices[symbol].to_records()

        return final_prices

    def add_prices(self,
                   symbol: str,
                   resolution: [str, int, float],
                   to: str = None,
                   start_date: typing.Union[str, float, int] = None,
                   stop_date: typing.Union[str, float, int] = None):
        """
        This is the user facing function for adding prices to the engine
        """
        start = None
        end = None

        resolution = int(time_interval_to_seconds(resolution))

        # Even if they specified start/end unevenly it will be overwritten with any to argument
        if to is not None:
            start = time.time() - time_interval_to_seconds(to)
            end = time.time()

        if start_date is not None:
            if isinstance(stop_date, (int, float)):
                start = start_date
            else:
                start_date = pd.to_datetime(start_date)
                epoch = dt.utcfromtimestamp(0)
                start = (start_date - epoch).total_seconds()

        if stop_date is not None:
            if isinstance(stop_date, (int, float)):
                end = stop_date
            else:
                end_date = pd.to_datetime(stop_date)
                epoch = dt.utcfromtimestamp(0)
                end = (end_date - epoch).total_seconds()

        # If start/ends are specified unevenly
        if (start_date is None and stop_date is not None) or (start_date is not None and stop_date is None):
            raise ValueError("Both start and end dates must be set or use the 'to' argument.")

        self.__add_prices(symbol, start, end, resolution)

    def add_custom_prices(self, price_reader: PriceReader):
        self.__price_readers.append(price_reader)

    def add_custom_events(self, event_reader: DataReader):
        self.__event_readers.append(event_reader)

    def add_tick_events(self, tick_reader: TickReader):
        self.__tick_readers.append(tick_reader)

    def __add_prices(self, symbol, start_time, end_time, resolution, save=False):
        # If it's not loaded then write it to the file
        # Add it as a new price
        self.__user_added_times.append({
            self.PriceIdentifiers.sandbox: True,
            self.PriceIdentifiers.symbol: symbol,
            self.PriceIdentifiers.resolution: resolution,
            self.PriceIdentifiers.epoch_start: start_time,
            self.PriceIdentifiers.epoch_stop: end_time
        })
        if save:
            self.queue_backtest_write = True

        self.__check_user_time_bounds(start_time, end_time, resolution)

    def __check_user_time_bounds(self, start_time, end_time, resolution):
        # This makes sure that we keep track of our bounds which is just generally kind of useful
        # Any time a new price event is added we check this
        if self.user_start is None:
            self.user_start = start_time
        else:
            if start_time < self.user_start:
                self.user_start = start_time

        if self.user_stop is None:
            self.user_stop = end_time
        else:
            if end_time > self.user_stop:
                self.user_stop = end_time

        # Now keep track of the smallest price event added
        if self.min_resolution is None:
            self.min_resolution = resolution
        else:
            if resolution < self.min_resolution:
                self.min_resolution = resolution

    def write_setting(self, key, value, save=False):
        """
        Write a setting to the .json preferences

        Args:
            key: Key under settings
            value: Value to set that settings key to
            save: Write this new setting to the file
        """
        self.preferences['settings'][key] = value
        if save:
            self.queue_backtest_write = True

    def __write_initial_price_values(self, account_dictionary):
        """
        Write in a new price dictionary for the paper trade exchange.
        """
        self.interface.override_local_account(account_dictionary)

    def format_account_data(self, interface: PaperTradeInterface, local_time) -> \
            typing.Tuple[typing.Dict[
                             typing.Union[
                                 str, typing.Any],
                             typing.Union[
                                 int, typing.Any]],
                         typing.Dict[
                             typing.Union[str,
                                          typing.Any],
                             typing.Union[
                                 int, typing.Any]]]:

        # This is done so that only traded assets are evaluated.
        true_available: dict = {}
        true_account: dict = {}
        for i in interface.traded_assets:
            # Grab the account status
            true_account[i] = interface.get_account(i)

        # Create an account total value
        value_total = 0

        no_trade_available = {}
        # No trade account total
        no_trade_value = 0

        # Save this up front so that it can be removed from the price calculation (it's always a value of 1 anyway)
        try:
            quote_value = true_account[self.quote_currency]['available'] + true_account[self.quote_currency]['hold']
        except KeyError as e:
            raise KeyError(f"Failed looking up {e}. Try changing your quote_account_value_in in backtest.json to be "
                           f"{e}.")
        try:
            del true_account[self.quote_currency]
        except KeyError:
            pass

        for i in list(true_account.keys()):
            # Funds on hold are still added
            true_available[i] = true_account[i]['available'] + true_account[i]['hold']
            no_trade_available[i] = self.initial_account[i]['available'] + self.initial_account[i]['hold']
            currency_pair = i

            # Convert to quote (this could be optimized a bit)
            is_stonks = interface.get_exchange_type() == 'alpaca'
            is_future = currency_pair.endswith('PERP')
            if not (is_stonks or is_future):
                currency_pair += '-'
                currency_pair += self.quote_currency

            # Get price at time
            try:
                price = interface.get_price(currency_pair)
            except KeyError:
                # Must be a currency we have no data for
                # This used to return zero but I believe all the cases where no data was avaiable have been elimated.
                raise KeyError(f"Failed to quote {currency_pair} because no downloaded data for that pair is available. "
                               f"Make sure to set \"quote_account_value_in\" in \"backtest.json\" to match the prices "
                               f"you are using. For example if you are trading \"USD-JPY\", set your quote value "
                               f"to \"JPY\". Currently it is set to {self.quote_currency}")

            # This is needed for futures apparently
            if is_future:
                value_total += price * abs(true_available[i])
                no_trade_value += price * abs(no_trade_available[i])
            else:
                # For stocks make sure not to use an absolute value
                value_total += price * true_available[i]
                no_trade_value += price * no_trade_available[i]

        # Make sure to add the time key in
        true_available['time'] = local_time
        no_trade_available['time'] = local_time

        value_total += quote_value
        true_available[self.quote_currency] = quote_value

        no_trade_value += self.initial_account[self.quote_currency][
                              'available'] + self.initial_account[self.quote_currency]['hold']

        true_available['Account Value (' + self.quote_currency + ')'] = value_total

        no_trade_available['Account Value (No Trades)'] = no_trade_value

        return true_available, no_trade_available

    def __account_was_used(self, column) -> bool:
        show_zero_delta = self.preferences['settings']['show_tickers_with_zero_delta']

        # Just check if it's in the traded assets or if the zero delta is enabled
        is_used = column in self.interface.traded_assets or 'Account Value (' + self.quote_currency + ')' == column

        # Return true if they are not the same or the setting is set to true
        output = is_used or show_zero_delta
        return output

    def __next_color(self):
        # This should be a generator, but it doesn't work without doing a foreach loop
        try:
            return next(self.__color_generator)
        except StopIteration:
            self.__color_generator = Category10_10.__iter__()
            return next(self.__color_generator)

    def advance_time_and_price_index(self):
        def handle_blankly_tick(type_: str, data):
            if type_ == 'tick':
                self.model.websocket_update(data)
            elif type_ == "funding_rate":
                self.interface.do_funding(data['symbol'], data['rate'])

        def run_events():
            # Ensure that we don't index error here
            events_length = len(self.events)
            # Make sure we don't crash at first
            if self.event_index >= events_length:
                return

            # Store the time because we need accurate time for the async stuff
            time_backup = self.time
            while self.events[self.event_index]['time'] < time_backup:
                # Set time to something different here
                event = self.events[self.event_index]
                self.time = event['time']
                if event['type'][0:11] != '__blankly__':
                    self.model.event(event['type'], event['data'])
                else:
                    handle_blankly_tick(event['type'][11:], event['data'])
                # Fired some event, go to the next one
                self.event_index += 1

                # Just check after doing that if we went outside the index
                if self.event_index >= events_length:
                    return

            self.time = time_backup

        # Now update the time to match
        self.interface.receive_time(self.time)

        for symbol in self.prices:
            # Make sure that each price column is at least at the current time

            # This just incrementing the price indexes until it's less than time and ensuring that
            #  it's less than the length of the price
            price_length = len(self.prices[symbol]) - 2
            while self.prices[symbol][self.price_indexes[symbol]]['time'] < self.time:
                if price_length >= self.price_indexes[symbol]:
                    self.price_indexes[symbol] += 1
                else:
                    self.model.has_data = False
                    break

            # Write this new price into the interface
            self.interface.receive_price(symbol, new_price=self.prices[symbol][
                self.price_indexes[symbol]][self.use_price])

        # Check has_data here also
        if self.time > self.user_stop:
            self.model.has_data = False

        run_events()

    def sleep(self, seconds: [int, float]):
        # Always evaluate limits
        self.interface.evaluate_limits()
        self.sleep_count += 1

        if self.show_progress:
            if self.sleep_count % 300 == 0:
                # Update the progress occasionally
                update_progress((self.time - self.user_start) / (self.user_stop - self.user_start))

        # Advance the time
        self.time += seconds
        # Refresh all the prices and times
        self.advance_time_and_price_index()

    def value_account(self) -> None:
        """
        Store the valuation for the account

        This is accessible by the user
        """
        # Don't do anything when live
        if not self.backtesting:
            return

        available_dict, no_trade_dict = self.format_account_data(self.interface, self.time)

        self.traded_account_values.append(available_dict)
        self.no_trade_account_values.append(no_trade_dict)

    # TODO this class should be constructed with a BacktestConfiguration object
    def run(self,
            args,
            exchange: ABCExchange,
            initial_account_values,
            backtest_settings_path: str = None,
            **kwargs) -> BacktestResult:
        """
        Setup
        """
        self.backtesting = True
        self.preferences = load_backtest_preferences(backtest_settings_path)
        # Write any dynamic arguments back into the backtest preferences
        for setting in kwargs:
            self.preferences['settings'][setting] = kwargs[setting]

        self.backtest_settings_path = backtest_settings_path
        self.show_progress = self.preferences['settings']['show_progress_during_backtest']

        if not exchange.get_type().endswith("paper_trade"):
            raise ValueError("Backtest controller was not constructed with a paper trade exchange object.")
        # Define the interface on run
        self.interface: PaperTradeInterface = exchange.get_interface()
        # This is where we begin logging the backtest time
        start_clock = time.time()

        # Figure out our traded assets here
        self.prices = self.sync_prices()
        # add funding rate events for futures trading
        if isinstance(self.interface, FuturesPaperTradeInterface):
            for symbol in self.prices:
                self.add_custom_events(FundingRateEventReader(symbol, self.user_start, self.user_stop, self.interface))
        # Now ensure all events are processed
        self.parse_events()
        for symbol in self.prices:
            base = get_base_asset(symbol)
            quote = get_quote_asset(symbol)
            if base not in self.interface.traded_assets:
                self.interface.traded_assets.append(base)
            if quote not in self.interface.traded_assets:
                self.interface.traded_assets.append(quote)

        # Write them in
        if initial_account_values is not None:
            self.__write_initial_price_values(initial_account_values)

        # Write our queued edits to the file
        if self.queue_backtest_write:
            write_backtest_preferences(self.preferences, self.backtest_settings_path)

        # TODO the preferences should really be reloaded here so that micro changes such as the quote currency reset
        #  don't need to happen for every single key type
        self.quote_currency = self.preferences['settings']['quote_account_value_in']

        # Get the symbol used for the benchmark
        benchmark_symbol = self.preferences["settings"]["benchmark_symbol"]

        if benchmark_symbol is not None:
            # Check locally for the data and add to price_cache if we do not have it
            self.add_prices(benchmark_symbol, start_date=self.user_start, stop_date=self.user_stop,
                            resolution=self.min_resolution)

        use_price = self.preferences['settings']['use_price']
        self.use_price = use_price

        for frame_symbol, price_list in self.prices.items():
            # This is a list of dictionaries
            frame = price_list  # type: list

            # Be sure to push these initial prices to the strategy
            try:
                self.interface.receive_price(frame_symbol, frame[0][use_price])
            except IndexError:
                def check_if_any_column_has_prices(price_dict: dict) -> bool:
                    """
                    In dictionary of symbols, check if at least one key has data
                    """
                    for j in price_dict:
                        # This handles the weird types
                        if len(price_dict[j]) == 0:
                            return False
                        if not price_dict[j].empty:
                            return True
                    return False

                if not check_if_any_column_has_prices(self.prices):
                    raise IndexError('No cached or downloaded data available. Try adding arguments such as to="1y" '
                                     'in the backtest command. If there should be data downloaded, try deleting your'
                                     f' ./price_caches folder. Also ensure that "{frame_symbol}" is spelled correctly.')
                else:
                    raise IndexError(f"Data for symbol {frame_symbol} is empty. Are you using a symbol that "
                                     f"is incompatible "
                                     f"with this exchange?")

            # Be sure to send in the initial time
            first_time = price_list['time'][0]
            self.interface.receive_time(first_time)
            self.price_indexes[frame_symbol] = 0

            # Find the first time in the list
            self.initial_time = copy.copy(self.user_start)
            self.interface.initial_time = self.initial_time

        if self.prices == {} and self.events == []:
            raise ValueError("No data given. "
                             "Try setting an argument such as to='1y' in the .backtest() command.\n"
                             "Example: strategy.backtest(to='1y')")

        """
        Begin backtesting
        """
        # Create this initial so that we can compare how our strategy performs
        self.initial_account = self.interface.get_account()

        # Initialize this before the callbacks, so it works in the initialization functions
        self.time = copy.copy(self.user_start)

        # Turn on backtesting immediately after setting the time
        self.interface.set_backtesting(True)

        # Re-evaluate the traded assets account
        # This is mainly used if the user has an account with some value that gets added in at the backtest point
        # This occurs after initialization so there has to be a function to test & re-evaluate that
        self.interface.evaluate_traded_account_assets()
        column_keys = list.copy(self.interface.traded_assets)

        # Comically if you don't include the quote at any point there will be an error
        if self.quote_currency not in column_keys:
            column_keys.append(self.quote_currency)
        # If they start a price event on something they don't own, this should also be included
        column_keys.append('time')

        cycle_status = pd.DataFrame(columns=column_keys)

        no_trade_cycle_status = pd.DataFrame(columns=column_keys)

        # Add an initial account row here
        if self.preferences['settings']['save_initial_account_value']:
            available_dict, no_trade_dict = self.format_account_data(self.interface, self.user_start)
            self.traded_account_values.append(available_dict)
            self.no_trade_account_values.append(no_trade_dict)

        print("\nBacktesting...")

        # Start the model here
        try:
            self.model.main(args)
            if self.show_progress:
                # If it finishes give it 100%
                update_progress(1)
        except Exception:
            traceback.print_exc()
        finally:
            self.model.teardown()

        # Reset time to indicate we are no longer in a backtest
        self.time = None

        # Push the accounts to the dataframe
        cycle_status = pd.concat([cycle_status, pd.DataFrame(self.traded_account_values)],
                                 ignore_index=True).sort_values(by=['time'])

        if len(cycle_status) == 0:
            raise RuntimeError("Empty result - no valid backtesting events occurred. Was there an error?.")

        no_trade_cycle_status = pd.concat([no_trade_cycle_status, pd.DataFrame(self.no_trade_account_values)],
                                          ignore_index=True).sort_values(by=['time'])

        def is_number(s):
            try:
                float(s)
                # Love how bools cast to a number
                return not isinstance(s, bool)
            except ValueError:
                return False

        history_and_returns: dict = {
            'history': cycle_status
        }
        metrics_indicators = {}
        user_callbacks = {}

        result_object = BacktestResult(history_and_returns, {
            'created': self.interface.paper_trade_orders,
            'limits_executed': self.interface.executed_orders,
            'limits_canceled': self.interface.canceled_orders,
            'executed_market_orders': self.interface.market_order_execution_details
        }, self.prices, self.initial_time, self.interface.time(), self.quote_currency, [])

        # If they set resampling we use resampling for everything
        resample_setting = self.preferences['settings']['resample_account_value_for_metrics']
        if isinstance(resample_setting, str) or is_number(resample_setting):
            resample_to = resample_setting
        else:
            info_print('Resampling value not set, defaulting to 1 day.')
            resample_to = '1d'

        interval_value = time_interval_to_seconds(resample_to)

        # This is where we run the actual resample
        resampled_account_data_frame = result_object.resample_account('Account Value (' + self.quote_currency + ')',
                                                                      interval_value)

        history_and_returns['resampled_account_value'] = resampled_account_data_frame

        returns = resampled_account_data_frame.copy(deep=True)

        # Default diff parameters should do it
        returns['value'] = returns['value'].pct_change()

        # Now write it to our dictionary
        history_and_returns['returns'] = returns

        # -----=====*****=====-----
        metrics_indicators['Compound Annual Growth Rate (%)'] = metrics.cagr(history_and_returns)
        try:
            metrics_indicators['Cumulative Returns (%)'] = metrics.cum_returns(history_and_returns)
        except ZeroDivisionError as e_:
            metrics_indicators['Cumulative Returns (%)'] = f'failed: {e_}'

        def attempt(math_callable: typing.Callable, dict_of_dataframes: dict, kwargs_: dict = None):
            try:
                if kwargs_ is None:
                    kwargs_ = {}
                result = math_callable(dict_of_dataframes, **kwargs_)
                if result == np.NAN:
                    result = None
                return result
            except (ZeroDivisionError, Exception) as e__:
                return f'failed: {e__}'

        risk_free_return_rate = self.preferences['settings']["risk_free_return_rate"]
        metrics_indicators['Max Drawdown (%)'] = attempt(metrics.max_drawdown, history_and_returns)
        metrics_indicators['Variance (%)'] = attempt(metrics.variance, history_and_returns,
                                                     {'trading_period': interval_value})
        metrics_indicators['Sortino Ratio'] = attempt(metrics.sortino, history_and_returns,
                                                      {'risk_free_rate': risk_free_return_rate,
                                                       'trading_period': interval_value})
        metrics_indicators['Sharpe Ratio'] = attempt(metrics.sharpe, history_and_returns,
                                                     {'risk_free_rate': risk_free_return_rate,
                                                      'trading_period': interval_value})
        metrics_indicators['Calmar Ratio'] = attempt(metrics.calmar, history_and_returns,
                                                     {'trading_period': interval_value})
        metrics_indicators['Volatility'] = attempt(metrics.volatility, history_and_returns,
                                                   {'trading_period': interval_value})
        metrics_indicators['Value-at-Risk'] = attempt(metrics.var, history_and_returns)
        metrics_indicators['Conditional Value-at-Risk'] = attempt(metrics.cvar, history_and_returns)

        # Add risk-free-return rate to dictionary
        metrics_indicators['Risk Free Return Rate'] = risk_free_return_rate
        # metrics_indicators['beta'] = attempt(metrics.beta, dataframes)
        # Add the interval value to dictionary
        metrics_indicators['Resampled Time'] = interval_value
        # -----=====*****=====-----

        # If a benchmark was requested, add it to the pd_prices frame
        if benchmark_symbol is not None:
            # Resample the benchmark results
            resampled_benchmark_value = result_object.resample_account(benchmark_symbol,
                                                                       interval_value,
                                                                       use_asset_history=True,
                                                                       use_price=use_price)

            # Push data into the dictionary for use by the metrics
            history_and_returns['benchmark_value'] = resampled_benchmark_value
            history_and_returns['benchmark_returns'] = resampled_benchmark_value.copy(deep=True)
            history_and_returns['benchmark_returns']['value'] = history_and_returns['benchmark_returns'][
                'value'].pct_change()

            # Calculate beta
            metrics_indicators['Beta'] = attempt(metrics.beta, history_and_returns,
                                                 {"trading_period": interval_value})

        # Remove NaN values here
        history_and_returns['resampled_account_value'] = history_and_returns['resampled_account_value']. \
            where(history_and_returns['resampled_account_value'].notnull(), None)

        # Remove NaN values on this one too
        history_and_returns['returns'] = history_and_returns['returns'].where(history_and_returns['returns'].notnull(),
                                                                              None)
        # Lastly remove Nan values in the metrics
        for symbol in metrics_indicators:
            if not isinstance(metrics_indicators[symbol], str) and np.isnan(metrics_indicators[symbol]):
                metrics_indicators[symbol] = None

        # Assign all these new values back to the result object
        result_object.history_and_returns = history_and_returns
        result_object.metrics = metrics_indicators
        result_object.user_callbacks = user_callbacks
        result_object.exchange = self.interface.get_exchange_type()

        figures = []
        # This modifies the platform result in place
        platform_result = format_platform_result(result_object)
        if self.preferences['settings']['GUI_output']:
            def internal_backtest_viewer():
                # for i in self.prices:
                #     result_index = cycle_status['time'].sub(i[0]).abs().idxmin()
                #     for i in cycle_status.iloc[result_index]:

                hover = HoverTool(
                    tooltips=[
                        ('value', '@value')
                    ],

                    # formatters={
                    #     'time': 'datetime',  # use 'datetime' formatter for 'date' field
                    #     '@{value}': 'printf',   # use 'printf' formatter for '@{adj close}' field
                    # },

                    # display a tooltip whenever the cursor is vertically in line with a glyph
                    mode='vline'
                )

                # Define a helper function to avoid repeating code
                def add_trace(self_, figure_, time__, data_, label):
                    source = ColumnDataSource(data=dict(
                        time=time__,
                        value=data_.values.tolist()
                    ))
                    figure_.step('time', 'value',
                                 source=source,
                                 line_width=2,
                                 color=self_.__next_color(),
                                 legend_label=label,
                                 mode="after")

                global_x_range = None

                time_ = [dt.fromtimestamp(ts) for ts in cycle_status['time']]

                for column in cycle_status:
                    if column != 'time' and self.__account_was_used(column):
                        p = figure(frame_width=900, frame_height=200, x_axis_type='datetime')
                        add_trace(self, p, time_, cycle_status[column], column)

                        # Add the no-trade line to the backtest
                        if column == 'Account Value (' + self.quote_currency + ')':
                            add_trace(self, p, time_, no_trade_cycle_status['Account Value (No Trades)'],
                                      'Account Value (No Trades)')

                            # Add the benchmark, if requested
                            if benchmark_symbol is not None:
                                # This normalizes the benchmark value
                                initial_account_value = cycle_status['Account Value (' +
                                                                     self.quote_currency + ')'].iloc[0]
                                benchmark_series = pd.Series(self.prices[benchmark_symbol][use_price])
                                initial_benchmark_value = benchmark_series.iloc[0]

                                # This multiplier brings the initial asset price to the initial account value
                                # initial_account_value = initial_benchmark_value * x
                                multiplier = initial_account_value / initial_benchmark_value

                                normalized_compare_series = benchmark_series.multiply(multiplier)
                                normalized_compare_time_series = self.prices[benchmark_symbol]['time']

                                # We need to also cast the time series that is needed to compare
                                # because it's only been done for the cycle status time
                                normalized_compare_time_series = [dt.fromtimestamp(ts) for ts in
                                                                  normalized_compare_time_series]
                                add_trace(self, p, normalized_compare_time_series,
                                          normalized_compare_series,
                                          f'Normalized Benchmark ({benchmark_symbol})')

                        p.add_tools(hover)

                        # Format graph
                        p.legend.location = "top_left"
                        p.legend.title = column
                        p.legend.title_text_font_style = "bold"
                        p.legend.title_text_font_size = "20px"
                        if global_x_range is None:
                            global_x_range = p.x_range
                        else:
                            p.x_range = global_x_range

                        figures.append(p)

                show(bokeh_columns(figures))
                # info_print(f'Make an account to take advantage of the platform backtest viewer: '
                #            f'https://app.blankly.finance/RETIe0J8EPSQz7wizoJX0OAFb8y1/62iIMVRKV7zkcpJysYlP/'
                #            f'75a0c190-4d8a-44e2-9310-c47d4d72b070/backtest')

            # This is where we end the backtesting time
            stop_clock = time.time()

            internal_backtest_viewer()
            # TODO this code does a good job uploading finished backtests to the platform. This should be fixed to
            #  allow configuration in the settings to reference any self hosted version of the platform
            # try:
            #     json_file = json.loads(open('./blankly.json').read())
            #     api_key = json_file['api_key']
            #     api_pass = json_file['api_pass']
            #     # Need this to generate the URL
            #     # Need this to know where to post to
            #     model_id = json_file['model_id']
            #
            #     requests.post(f'https://events.blankly.finance/v1/backtest/result', json=platform_result, headers={
            #         'api_key': api_key,
            #         'api_pass': api_pass,
            #         'model_id': model_id
            #     })
            #
            #     requests.post(f'https://events.blankly.finance/v1/backtest/status', json={
            #         'successful': True,
            #         'status_summary': 'Completed',
            #         'status_details': '',
            #         'time_elapsed': stop_clock - start_clock,
            #         'backtest_id': platform_result['backtest_id']
            #     }, headers={
            #         'api_key': api_key,
            #         'api_pass': api_pass,
            #         'model_id': model_id
            #     })
            #
            #     import webbrowser
            #
            #     link = f'https://app.blankly.finance/{api_key}/{model_id}/{platform_result["backtest_id"]}' \
            #            f'/backtest'
            #     webbrowser.open(
            #         link
            #     )
            #     info_print(f'View your backtest here: {link}')
            # except (FileNotFoundError, KeyError):
            #     internal_backtest_viewer()

        # Finally, write the figures in
        result_object.figures = figures

        self.interface.set_backtesting(False)
        self.backtesting = False

        # Export to the platform here
        blankly.reporter.export_backtest_result(platform_result)

        return result_object



"""
    A similar state management to StrategyState but optimized for evaluation & reporting
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time

from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface as Interface
from blankly.utils.utils import AttributeDict, format_with_new_line


class ScreenerState:
    interface: Interface
    variables: AttributeDict
    symbols: list

    def __init__(self, screener):
        """
        This is a ScreenerState object - a simplified version of StrategyState

        Args:
            screener: Construct with a screener object to allow interaction
        """

        self.screener = screener
        self.variables = AttributeDict({})
        self.symbols = screener.symbols

    @property
    def interface(self) -> Interface:
        """
        Get the interface object to interact with the exchange
        """
        return self.screener.interface

    @property
    def time(self) -> float:
        """
        Get the current time. This will only the current time because it is not in a backtesting framework
        """
        return time.time()

    def notify(self, message=None) -> None:
        """
        Send the formatted results as an email
        """
        self.screener.notify(message)

    def __str__(self):
        output = ""
        output = format_with_new_line(output, 'Time: ', self.time)
        output = format_with_new_line(output, 'Symbols: ', self.symbols)
        return format_with_new_line(output, 'Variables: ', self.variables)



"""
    Interface definition for homogenous calls
    Copyright (C) 2021 Blankly Finance

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import time
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
import blankly.utils.utils as utils
import copy
from typing import List

from blankly.exchanges.orders.stop_loss import StopLossOrder
from blankly.exchanges.orders.take_profit import TakeProfitOrder


class FTXInterface(ExchangeInterface):

    # note, FTX has no sandbox mode
    def __init__(self, exchange_name, authenticated_api: FTXAPI):
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[15, 60, 300, 900,
                                                                              3600, 14400, 86400, 86400])

    def init_exchange(self):
        """
            Attempts an API call. If API key is invalid, this 
            method will throw another error specifying that it 
            was due to invalid FTX API keys 
        """
        try:
            self.get_calls().get_account_info()
        except Exception as e:
            if str(e) == "Not logged in: Invalid API key":
                raise LookupError("Unable to connect to FTX US: Invalid API Key")
            else:
                raise e

    """
    needed:
    ---
    'get_products': [
                ["symbol", str],
                ["base_asset", str],
                ["quote_asset", str],
                ["base_min_size", float],
                ["base_max_size", float],
                ["base_increment", float]
            ],

    provided:
    ---
    {
            "name": "ABNB/USD",
            "enabled": true,
            "postOnly": false,
            "priceIncrement": 0.005, //NOTE: api call returns both price and size increment. base_increment is currently
             set to size, however this may not be the expected result
            "sizeIncrement": 0.025,
            "minProvideSize": 0.025,
            "last": 169.09,
            "bid": 170.68,
            "ask": 171.205,
            "price": 170.68,
            "type": "spot",
            "baseCurrency": "ABNB",
            "quoteCurrency": "USD",
            "underlying": null,
            "restricted": true,
            "tokenizedEquity": true,
            "highLeverageFeeExempt": true,
            "change1h": -0.003590297440088736,
            "change24h": 0.00014649438926489115,
            "changeBod": -0.00367754363434709,
            "quoteVolume24h": 12.76575,
            "volumeUsd24h": 12.76575
        },
    """

    # only includes markets of type "spot" (i.e. excludes futures)
    def get_products(self) -> list:

        needed = self.needed['get_products']

        products: List[dict] = self.get_calls().list_markets()

        end_products = []

        for index, product in enumerate(products):
            if product['type'] == "spot":
                product['symbol'] = utils.to_blankly_symbol(product.pop('name'), 'ftx')
                product['base_asset'] = product.pop('baseCurrency')
                product['quote_asset'] = product.pop('quoteCurrency')
                product['base_min_size'] = product.pop('minProvideSize')
                product['base_max_size'] = 99999999
                product['base_increment'] = product.pop('sizeIncrement')

                product = utils.isolate_specific(needed, product)

                end_products.append(product)

            else:
                continue

        return end_products

    @utils.enforce_base_asset
    def get_account(self, symbol: str = None) -> utils.AttributeDict:
        """
        Get all assets in an account, or sort by symbol/account_id
        Args:
            symbol (Optional): Filter by particular symbol

            These arguments are mutually exclusive

        TODO add return example
        """

        symbol = super().get_account(symbol=symbol)

        needed = self.needed['get_account']
        parsed_dictionary = utils.AttributeDict({})
        balances: List[dict] = self.get_calls().get_balances()

        if symbol is not None:
            for account in balances:
                if account["coin"] == symbol:
                    account['hold'] = account['total'] - account['free']
                    account['currency'] = account.pop('coin')
                    account['available'] = account.pop('free')
                    parsed_value = utils.isolate_specific(needed, account)
                    requested_account = utils.AttributeDict({
                        'available': parsed_value['available'],
                        'hold': parsed_value['hold']
                    })
                    return requested_account
            raise ValueError("Symbol not found")

        else:
            for account in balances:
                account['hold'] = account['total'] - account['free']
                account['currency'] = account.pop('coin')
                account['available'] = account.pop('free')
                parsed_dictionary[account['currency']] = utils.AttributeDict({
                    'available': float(account['available']),
                    'hold': float(account['hold'])
                })

            return utils.add_all_products(parsed_dictionary, self.get_products())

    """
    needed:

    'market_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["size", float],
                ["status", str],
                ["type", str],
                ["side", str]
            ],

    response: 

    {
    "createdAt": "2019-03-05T09:56:55.728933+00:00",
    "filledSize": 0,
    "future": "XRP-PERP",
    "id": 9596912,
    "market": "XRP-PERP",
    "price": 0.306525,
    "remainingSize": 31431,
    "side": "sell",
    "size": 31431,
    "status": "open",
    "type": "limit",
    "reduceOnly": false,
    "ioc": false,
    "postOnly": false,
    "clientId": null,
    }
    """

    @utils.order_protection
    def market_order(self, symbol: str, side: str, size: float) -> MarketOrder:
        """
        Used for buying or selling market orders
        Args:
            symbol: asset to buy
            side: buy/sell
            size: desired amount of base asset to use
        """
        needed = self.needed['market_order']
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))

        # TODO pretty sure this needs to convert the symbol to an 'FTX' symbol before sending it off to the api
        response = self.get_calls().place_order(symbol, side, None, size, order_type="market")

        response = self._fix_response(needed, response, False)

        order = utils.build_order_info(0, side, size, symbol, 'market')

        return MarketOrder(order, response, self)

    """
    needed

    'limit_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["price", float],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
                ],

    response

    {
    "success": true,
    "result": {
        "createdAt": "2019-03-05T09:56:55.728933+00:00",
        "filledSize": 0,
        "future": "XRP-PERP",
        "id": 9596912,
        "market": "XRP-PERP",
        "price": 0.306525,
        "remainingSize": 31431,
        "side": "sell",
        "size": 31431,
        "status": "open",
        "type": "limit",
        "reduceOnly": false,
        "ioc": false,
        "postOnly": false,
        "clientId": null,
        }
    }
    """

    @utils.order_protection
    def limit_order(self,
                    symbol: str,
                    side: str,
                    price: float,
                    size: float) -> LimitOrder:
        """
        Used for buying or selling limit orders
        Args:
            symbol: asset to buy
            side: buy/sell
            price: price to set limit order
            size: amount of asset (like BTC) for the limit to be valued
        """
        needed = self.needed['take_profit']
        response = self.get_calls().place_order(symbol, side, price, size, order_type="limit")
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = utils.build_order_info(price, side, size, symbol, 'take_profit')

        response = self._fix_response(needed, response)

        return LimitOrder(order, response, self)

    @staticmethod
    def _fix_response(needed, response, set_tif=True):
        response["symbol"] = utils.to_blankly_symbol(response.pop("market"), 'ftx')
        response["created_at"] = utils.epoch_from_iso8601(response.pop("createdAt"))
        if set_tif:
            response["time_in_force"] = "GTC"
        response = utils.isolate_specific(needed, response)
        return response

    @utils.order_protection
    def stop_loss_order(self,
                        symbol: str,
                        price: float,
                        size: float) -> StopLossOrder:
        """
        Used for placing a stop-loss order
        Args:
            symbol: asset to buy
            price: price to sell at
            size: amount of asset (like BTC)
        """
        needed = self.needed['stop_loss']
        side = 'sell'
        response = self.get_calls().place_conditional_order(symbol, side, size, order_type="stop",
                                                            trigger_price=price)

        order = utils.build_order_info(price, side, size, symbol, 'stop_loss')

        response = self._fix_response(needed, response)

        return StopLossOrder(order, response, self)

    @utils.order_protection
    def take_profit_order(self,
                          symbol: str,
                          price: float,
                          size: float) -> TakeProfitOrder:
        """
        Used for placing a take-profit order
        Args:
            symbol: asset to buy
            price: price to sell at
            size: amount of asset (like BTC)
        """
        needed = self.needed['limit_order']
        side = 'sell'
        response = self.get_calls().place_conditional_order(symbol, side, size, order_type="takeProfit",
                                                            trigger_price=price)

        order = utils.build_order_info(price, side, size, symbol, 'take_profit')

        response = self._fix_response(needed, response)

        return TakeProfitOrder(order, response, self)

    def cancel_order(self, symbol: str, order_id: str) -> dict:
        """
        Cancel an order on a particular symbol & order id

        Args:
            symbol: This is the asset id that the order is under
            order_id: The unique ID of the order.

        """
        self.get_calls().cancel_order(order_id)
        return {'order_id': order_id}

    """
    [
        {
        "createdAt": "2019-03-05T09:56:55.728933+00:00",
        "filledSize": 10,
        "future": "XRP-PERP",
        "id": 9596912,
        "market": "XRP-PERP",
        "price": 0.306525,
        "avgFillPrice": 0.306526,
        "remainingSize": 31421,
        "side": "sell",
        "size": 31431,
        "status": "open",
        "type": "limit",
        "reduceOnly": false,
        "ioc": false,
        "postOnly": false,
        "clientId": null
        },
        {
            ...
        },
        ...
        {
            ...
        }

    ]
    """

    def get_open_orders(self, symbol: str = None) -> list:
        """
        List open orders.
        Args:
            symbol (optional) (str): Asset such as BTC-USD
        """
        response = self.get_calls().get_open_orders(symbol)
        response_needed_fulfilled = []
        if len(response) == 0:
            return []

        for open_order in response:
            open_order['symbol'] = utils.to_blankly_symbol(open_order.pop('market'), 'ftx')
            if open_order["type"] == "limit":
                needed = self.choose_order_specificity("limit")
                open_order['time_in_force'] = "GTC"
            elif open_order["type"] == "market":
                needed = self.choose_order_specificity("market")
            else:
                print(f"Order type {open_order['type']} is not supported.")
                continue

            open_order['created_at'] = utils.epoch_from_iso8601(open_order.pop('createdAt'))
            open_order = utils.isolate_specific(needed, open_order)
            response_needed_fulfilled.append(open_order)

        return response

    """
    {
        "createdAt": "2019-03-05T09:56:55.728933+00:00",
        "filledSize": 10,
        "future": "XRP-PERP",
        "id": 9596912,
        "market": "XRP-PERP",
        "price": 0.306525,
        "avgFillPrice": 0.306526,
        "remainingSize": 31421,
        "side": "sell",
        "size": 31431,
        "status": "open",
        "type": "limit",
        "reduceOnly": false,
        "ioc": false,
        "postOnly": false,
        "clientId": "your_client_order_id"
    }
    """

    def get_order(self, symbol: str, order_id: str) -> dict:
        """
        Get a certain order
        Args:
            symbol: Asset that the order is under
            order_id: The unique ID of the order.
        """
        response = self.get_calls().get_order_by_id(order_id)

        needed = self.choose_order_specificity(response["type"])

        response['symbol'] = utils.to_blankly_symbol(response.pop('market'), 'ftx')
        response['created_at'] = utils.epoch_from_iso8601(response.pop('createdAt'))

        if response["type"] == "limit":
            response['time_in_force'] = "GTC"

        response = utils.isolate_specific(needed, response)
        return response

    def get_fees(self, symbol) -> dict:
        """
        Get market fees

        lot of unnecessary info in exchange-specific section

        TODO: remove everything truly irrelevant to fees
        """

        needed = self.needed['get_fees']
        account_info = self.get_calls().get_account_info()
        account_info['maker_fee_rate'] = account_info.pop('makerFee')
        account_info['taker_fee_rate'] = account_info.pop('takerFee')
        account_info = utils.isolate_specific(needed, account_info)

        return account_info

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        """
        Returns the product history from an exchange
        Args:
            symbol: blankly product ID format (BTC-USD)
            epoch_start: Time to begin download
            epoch_stop: Time to stop download
            resolution: Resolution in seconds between tick (ex: 60 = 1 per minute)
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
        """
        return self.ftx_product_history(self.get_calls(), symbol, epoch_start, epoch_stop, resolution)

    @staticmethod
    def ftx_product_history(api, symbol, epoch_start, epoch_stop, resolution):
        # TODO this is probably the wrong place to put this code
        # shared binance history code needs to be refactored as well
        symbol = symbol.replace("-", "/")

        # epoch_start, epoch_stop = super().get_product_history(symbol, epoch_start, epoch_stop, resolution)
        epoch_start = utils.convert_epochs(epoch_start)
        epoch_stop = utils.convert_epochs(epoch_stop)

        accepted_grans = [15, 60, 300, 900, 3600, 14400, 86400]

        # ftx accepts the above resolutions plus any multiple of 86400 up to 30 * 86400
        for i in range(2, 31):
            accepted_grans.append(i * 86400)

        if resolution not in accepted_grans:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = accepted_grans[min(range(len(accepted_grans)),
                                            key=lambda j: abs(accepted_grans[j] - resolution))]

        # Figure out how many points are needed
        need = int((epoch_stop - epoch_start) / resolution)
        initial_need = need
        window_open = epoch_start
        history = []
        # Iterate while it's more than max
        while need > 1500:
            # Close is always 1500 points ahead
            window_close = window_open + 1500 * resolution

            response = api.get_product_history(symbol, window_open, window_close, resolution)

            history = history + response

            window_open = window_close
            need -= 1500
            time.sleep(.2)
            utils.update_progress((initial_need - need) / initial_need)

        # Fill the remainder
        window_close = epoch_stop
        response = api.get_product_history(symbol, window_open, window_close, resolution)
        history_block = history + response
        # print(history_block)
        history_block.sort(key=lambda x: x["time"])

        df = pd.DataFrame(history_block, columns=['time', 'low', 'high', 'open', 'close', 'volume'])
        # df[['time']] = df[['time']].astype(int)

        df["time"] = df["time"].apply(lambda x: x / 1000)
        df[['time']] = df[['time']].astype(int)

        df[['low', 'high', 'open', 'close', 'volume']] = df[['low', 'high', 'open', 'close', 'volume']].astype(float)

        return df

    def get_order_filter(self, symbol: str):
        """
        Find order limits for the exchange
        Args:
            symbol: The asset such as (BTC-USD, or MSFT)

            {
                "name": "BTC/USD", <---needed
                "enabled": true,
                "postOnly": false,
                "priceIncrement": 1.0, <------ needed
                "sizeIncrement": 0.0001, <------ needed
                "minProvideSize": 0.0001, <------ needed
                "last": 47534.0,
                "bid": 47536.0,
                "ask": 47540.0,
                "price": 47536.0,
                "type": "spot",
                "baseCurrency": "BTC", <------ needed
                "quoteCurrency": "USD", <------ needed
                "underlying": null,
                "restricted": false,
                "highLeverageFeeExempt": true,
                "change1h": -0.007184628237259816,
                "change24h": -0.003438155136268344,
                "changeBod": 0.02373260972563208,
                "quoteVolume24h": 57477234.2424,
                "volumeUsd24h": 57477234.2424
            }

        """
        if "-" in symbol:
            symbol = symbol.replace("-", "/")

        market_info = self.get_calls().get_market(symbol)

        exchange_specific_keys = copy.deepcopy(market_info)

        keys_to_remove = ["name", "baseCurrency", "quoteCurrency", "minProvideSize", "sizeIncrement", "priceIncrement",
                          "price"]

        for key in keys_to_remove:
            exchange_specific_keys.pop(key)

        return {
            "symbol": utils.to_blankly_symbol(market_info["name"], 'ftx'),  # make function in utils (or static)
            # to switch slash to dash
            "base_asset": market_info["baseCurrency"],
            "quote_asset": market_info["quoteCurrency"],
            "max_orders": 99999999,
            "limit_order": {
                "base_min_size": market_info["minProvideSize"],
                "base_max_size": 99999999,
                "base_increment": market_info["sizeIncrement"],
                "price_increment": market_info["priceIncrement"],
                "min_price": .75 * market_info["price"],
                "max_price": 1.25 * market_info["price"]
            },
            "market_order": {
                "base_min_size": market_info["minProvideSize"],
                "base_max_size": 99999999,
                "base_increment": market_info["sizeIncrement"],
                "quote_increment": market_info["priceIncrement"],
                "buy": {
                    "min_funds": market_info["sizeIncrement"],
                    "max_funds": market_info["sizeIncrement"]
                },
                "sell": {
                    "min_funds": market_info["sizeIncrement"],
                    "max_funds": market_info["sizeIncrement"]
                },

            },
            "exchange_specific": exchange_specific_keys
        }

    def get_price(self, symbol: str) -> float:
        """
        Returns just the price of a symbol.
        Args:
            symbol: The asset such as (BTC-USD, or MSFT)
        """
        return float(self.get_calls().get_market(symbol)['price'])
